/* Generated by Cython 0.28.5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "thriftrw/protocol/_endian.h"
        ], 
        "include_dirs": [
            "./thriftrw/protocol"
        ], 
        "name": "thriftrw.protocol.binary", 
        "sources": [
            "thriftrw/protocol/binary.pyx"
        ]
    }, 
    "module_name": "thriftrw.protocol.binary"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_5"
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__thriftrw__protocol__binary
#define __PYX_HAVE_API__thriftrw__protocol__binary
/* Early includes */
#include <stdint.h>
#include "_endian.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "thriftrw/protocol/binary.pyx",
  "stringsource",
  "thriftrw/wire/value.pxd",
  "thriftrw/wire/message.pxd",
  "thriftrw/_buffer.pxd",
  "thriftrw/protocol/core.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor;
struct __pyx_obj_8thriftrw_4wire_5value__ValueVisitorArgs;
struct __pyx_obj_8thriftrw_4wire_5value_Value;
struct __pyx_obj_8thriftrw_4wire_5value_BoolValue;
struct __pyx_obj_8thriftrw_4wire_5value_ByteValue;
struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue;
struct __pyx_obj_8thriftrw_4wire_5value_I16Value;
struct __pyx_obj_8thriftrw_4wire_5value_I32Value;
struct __pyx_obj_8thriftrw_4wire_5value_I64Value;
struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue;
struct __pyx_obj_8thriftrw_4wire_5value_FieldValue;
struct __pyx_obj_8thriftrw_4wire_5value_StructValue;
struct __pyx_obj_8thriftrw_4wire_5value_MapItem;
struct __pyx_obj_8thriftrw_4wire_5value_MapValue;
struct __pyx_obj_8thriftrw_4wire_5value_SetValue;
struct __pyx_obj_8thriftrw_4wire_5value_ListValue;
struct __pyx_obj_8thriftrw_4wire_7message_Message;
struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer;
struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer;
struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader;
struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter;
struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader;
struct __pyx_obj_8thriftrw_8protocol_4core_Protocol;
struct __pyx_obj_8thriftrw_8protocol_4core__ValueWriter;
struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol;
struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader;
struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter;
struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py;

/* "ttype.pxd":24
 * 
 * 
 * cpdef enum TType:             # <<<<<<<<<<<<<<
 *     BOOL = 2
 *     BYTE = 3
 */
enum __pyx_t_8thriftrw_4wire_5ttype_TType {
  __pyx_e_8thriftrw_4wire_5ttype_BOOL = 2,
  __pyx_e_8thriftrw_4wire_5ttype_BYTE = 3,
  __pyx_e_8thriftrw_4wire_5ttype_DOUBLE = 4,
  __pyx_e_8thriftrw_4wire_5ttype_I16 = 6,
  __pyx_e_8thriftrw_4wire_5ttype_I32 = 8,
  __pyx_e_8thriftrw_4wire_5ttype_I64 = 10,
  __pyx_e_8thriftrw_4wire_5ttype_BINARY = 11,
  __pyx_e_8thriftrw_4wire_5ttype_STRUCT = 12,
  __pyx_e_8thriftrw_4wire_5ttype_MAP = 13,
  __pyx_e_8thriftrw_4wire_5ttype_SET = 14,
  __pyx_e_8thriftrw_4wire_5ttype_LIST = 15
};
struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader;
struct __pyx_t_8thriftrw_8protocol_4core_MapHeader;
struct __pyx_t_8thriftrw_8protocol_4core_SetHeader;
struct __pyx_t_8thriftrw_8protocol_4core_ListHeader;

/* "core.pxd":36
 * 
 * 
 * cdef struct FieldHeader:             # <<<<<<<<<<<<<<
 *     int8_t type
 *     int16_t id
 */
struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader {
  int8_t type;
  int16_t id;
};

/* "core.pxd":41
 * 
 * 
 * cdef struct MapHeader:             # <<<<<<<<<<<<<<
 *     int8_t ktype
 *     int8_t vtype
 */
struct __pyx_t_8thriftrw_8protocol_4core_MapHeader {
  int8_t ktype;
  int8_t vtype;
  int32_t size;
};

/* "core.pxd":47
 * 
 * 
 * cdef struct SetHeader:             # <<<<<<<<<<<<<<
 *     int8_t type
 *     int32_t size
 */
struct __pyx_t_8thriftrw_8protocol_4core_SetHeader {
  int8_t type;
  int32_t size;
};

/* "core.pxd":52
 * 
 * 
 * cdef struct ListHeader:             # <<<<<<<<<<<<<<
 *     int8_t type
 *     int32_t size
 */
struct __pyx_t_8thriftrw_8protocol_4core_ListHeader {
  int8_t type;
  int32_t size;
};

/* "thriftrw/wire/value.pxd":31
 * 
 * 
 * cdef class ValueVisitor(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object visit_bool(self, bint value)
 */
struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_ValueVisitor *__pyx_vtab;
};


/* "thriftrw/wire/value.pxd":56
 * 
 * 
 * cdef class _ValueVisitorArgs(ValueVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8thriftrw_4wire_5value__ValueVisitorArgs {
  struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor __pyx_base;
};


/* "thriftrw/wire/value.pxd":60
 * 
 * 
 * cdef class Value(object):             # <<<<<<<<<<<<<<
 * 
 *     cpdef object apply(Value self, ValueVisitor visitor)
 */
struct __pyx_obj_8thriftrw_4wire_5value_Value {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value *__pyx_vtab;
};


/* "thriftrw/wire/value.pxd":65
 * 
 * 
 * cdef class BoolValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly bint value
 * 
 */
struct __pyx_obj_8thriftrw_4wire_5value_BoolValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int value;
};


/* "thriftrw/wire/value.pxd":69
 * 
 * 
 * cdef class ByteValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t value
 * 
 */
struct __pyx_obj_8thriftrw_4wire_5value_ByteValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int8_t value;
};


/* "thriftrw/wire/value.pxd":73
 * 
 * 
 * cdef class DoubleValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly double value
 * 
 */
struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  double value;
};


/* "thriftrw/wire/value.pxd":77
 * 
 * 
 * cdef class I16Value(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int16_t value
 * 
 */
struct __pyx_obj_8thriftrw_4wire_5value_I16Value {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int16_t value;
};


/* "thriftrw/wire/value.pxd":81
 * 
 * 
 * cdef class I32Value(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int32_t value
 * 
 */
struct __pyx_obj_8thriftrw_4wire_5value_I32Value {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int32_t value;
};


/* "thriftrw/wire/value.pxd":85
 * 
 * 
 * cdef class I64Value(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t value
 * 
 */
struct __pyx_obj_8thriftrw_4wire_5value_I64Value {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int64_t value;
};


/* "thriftrw/wire/value.pxd":89
 * 
 * 
 * cdef class BinaryValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly bytes value
 *     # TODO change to char* once BinaryProtocol knows how to write that.
 */
struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  PyObject *value;
};


/* "thriftrw/wire/value.pxd":94
 * 
 * 
 * cdef class FieldValue(object):             # <<<<<<<<<<<<<<
 *     cdef readonly int16_t id
 *     cdef readonly int8_t ttype
 */
struct __pyx_obj_8thriftrw_4wire_5value_FieldValue {
  PyObject_HEAD
  int16_t id;
  int8_t ttype;
  PyObject *value;
};


/* "thriftrw/wire/value.pxd":100
 * 
 * 
 * cdef class StructValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly list fields
 *     cdef readonly dict _index
 */
struct __pyx_obj_8thriftrw_4wire_5value_StructValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  PyObject *fields;
  PyObject *_index;
};


/* "thriftrw/wire/value.pxd":105
 * 
 * 
 * cdef class MapItem(object):             # <<<<<<<<<<<<<<
 *     cdef readonly object key
 *     cdef readonly object value
 */
struct __pyx_obj_8thriftrw_4wire_5value_MapItem {
  PyObject_HEAD
  PyObject *key;
  PyObject *value;
};


/* "thriftrw/wire/value.pxd":110
 * 
 * 
 * cdef class MapValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t key_ttype
 *     cdef readonly int8_t value_ttype
 */
struct __pyx_obj_8thriftrw_4wire_5value_MapValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int8_t key_ttype;
  int8_t value_ttype;
  PyObject *pairs;
};


/* "thriftrw/wire/value.pxd":116
 * 
 * 
 * cdef class SetValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t value_ttype
 *     cdef readonly list values
 */
struct __pyx_obj_8thriftrw_4wire_5value_SetValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int8_t value_ttype;
  PyObject *values;
};


/* "thriftrw/wire/value.pxd":121
 * 
 * 
 * cdef class ListValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t value_ttype
 *     cdef readonly list values
 */
struct __pyx_obj_8thriftrw_4wire_5value_ListValue {
  struct __pyx_obj_8thriftrw_4wire_5value_Value __pyx_base;
  int8_t value_ttype;
  PyObject *values;
};


/* "thriftrw/wire/message.pxd":26
 * 
 * 
 * cdef class Message(object):             # <<<<<<<<<<<<<<
 *     cdef public bytes name
 *     cdef public int32_t seqid
 */
struct __pyx_obj_8thriftrw_4wire_7message_Message {
  PyObject_HEAD
  PyObject *name;
  int32_t seqid;
  int message_type;
  PyObject *body;
};


/* "thriftrw/_buffer.pxd":24
 * 
 * 
 * cdef class ReadBuffer(object):             # <<<<<<<<<<<<<<
 *     cdef char* data
 *     cdef int offset, length
 */
struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *__pyx_vtab;
  char *data;
  int offset;
  int length;
  PyObject *_data;
};


/* "thriftrw/_buffer.pxd":39
 * 
 * 
 * cdef class WriteBuffer(object):             # <<<<<<<<<<<<<<
 *     cdef char* data
 *     cdef readonly int length, capacity
 */
struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_7_buffer_WriteBuffer *__pyx_vtab;
  char *data;
  int length;
  int capacity;
};


/* "core.pxd":57
 * 
 * 
 * cdef class MessageHeader(object):             # <<<<<<<<<<<<<<
 *     cdef readonly bytes name
 *     cdef readonly int8_t type
 */
struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader {
  PyObject_HEAD
  PyObject *name;
  int8_t type;
  int32_t seqid;
};


/* "core.pxd":63
 * 
 * 
 * cdef class ProtocolWriter(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_value(self, Value value) except *
 */
struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolWriter *__pyx_vtab;
};


/* "core.pxd":101
 * 
 * 
 * cdef class ProtocolReader:             # <<<<<<<<<<<<<<
 * 
 *     # Skip
 */
struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolReader *__pyx_vtab;
};


/* "core.pxd":146
 * 
 * 
 * cdef class Protocol(object):             # <<<<<<<<<<<<<<
 *     cpdef ProtocolWriter writer(self, WriteBuffer buff)
 * 
 */
struct __pyx_obj_8thriftrw_8protocol_4core_Protocol {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_8protocol_4core_Protocol *__pyx_vtab;
};


/* "core.pxd":160
 * 
 * 
 * cdef class _ValueWriter(ValueVisitor):             # <<<<<<<<<<<<<<
 *     cdef ProtocolWriter writer
 */
struct __pyx_obj_8thriftrw_8protocol_4core__ValueWriter {
  struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor __pyx_base;
  struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *writer;
};


/* "thriftrw/protocol/binary.pxd":61
 * 
 * 
 * cdef class BinaryProtocol(Protocol):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol {
  struct __pyx_obj_8thriftrw_8protocol_4core_Protocol __pyx_base;
};


/* "thriftrw/protocol/binary.pxd":64
 *     pass
 * 
 * cdef class BinaryProtocolReader(ProtocolReader):             # <<<<<<<<<<<<<<
 *     cdef ReadBuffer reader
 * 
 */
struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader {
  struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader __pyx_base;
  struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *reader;
};


/* "thriftrw/protocol/binary.pxd":75
 * 
 * 
 * cdef class BinaryProtocolWriter(ProtocolWriter):             # <<<<<<<<<<<<<<
 *     cdef WriteBuffer writer
 * 
 */
struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter {
  struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter __pyx_base;
  struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *writer;
};


/* "thriftrw/protocol/binary.pxd":81
 * 
 * 
 * cdef class _OldBinaryProtocolReader(object):             # <<<<<<<<<<<<<<
 *     cdef ReadBuffer reader
 * 
 */
struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_vtab;
  struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *reader;
};


/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(BoolValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 */
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_I16Value *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_I32Value *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_I64Value *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_StructValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_MapValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_SetValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};

struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py {
  PyObject_HEAD
  struct __pyx_obj_8thriftrw_4wire_5value_ListValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};



/* "thriftrw/wire/value.pxd":31
 * 
 * 
 * cdef class ValueVisitor(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object visit_bool(self, bint value)
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_ValueVisitor {
  PyObject *(*visit_bool)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int);
  PyObject *(*visit_byte)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int8_t);
  PyObject *(*visit_double)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, double);
  PyObject *(*visit_i16)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int16_t);
  PyObject *(*visit_i32)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int32_t);
  PyObject *(*visit_i64)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int64_t);
  PyObject *(*visit_binary)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, PyObject *);
  PyObject *(*visit_struct)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, PyObject *);
  PyObject *(*visit_map)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int8_t, int8_t, PyObject *);
  PyObject *(*visit_set)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int8_t, PyObject *);
  PyObject *(*visit_list)(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int8_t, PyObject *);
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_ValueVisitor *__pyx_vtabptr_8thriftrw_4wire_5value_ValueVisitor;


/* "thriftrw/wire/value.pxd":56
 * 
 * 
 * cdef class _ValueVisitorArgs(ValueVisitor):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value__ValueVisitorArgs {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_ValueVisitor __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value__ValueVisitorArgs *__pyx_vtabptr_8thriftrw_4wire_5value__ValueVisitorArgs;


/* "thriftrw/wire/value.pxd":60
 * 
 * 
 * cdef class Value(object):             # <<<<<<<<<<<<<<
 * 
 *     cpdef object apply(Value self, ValueVisitor visitor)
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value {
  PyObject *(*apply)(struct __pyx_obj_8thriftrw_4wire_5value_Value *, struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value *__pyx_vtabptr_8thriftrw_4wire_5value_Value;


/* "thriftrw/wire/value.pxd":65
 * 
 * 
 * cdef class BoolValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly bint value
 * 
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_BoolValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_BoolValue *__pyx_vtabptr_8thriftrw_4wire_5value_BoolValue;


/* "thriftrw/wire/value.pxd":69
 * 
 * 
 * cdef class ByteValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t value
 * 
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_ByteValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_ByteValue *__pyx_vtabptr_8thriftrw_4wire_5value_ByteValue;


/* "thriftrw/wire/value.pxd":73
 * 
 * 
 * cdef class DoubleValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly double value
 * 
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_DoubleValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_DoubleValue *__pyx_vtabptr_8thriftrw_4wire_5value_DoubleValue;


/* "thriftrw/wire/value.pxd":77
 * 
 * 
 * cdef class I16Value(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int16_t value
 * 
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_I16Value {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_I16Value *__pyx_vtabptr_8thriftrw_4wire_5value_I16Value;


/* "thriftrw/wire/value.pxd":81
 * 
 * 
 * cdef class I32Value(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int32_t value
 * 
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_I32Value {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_I32Value *__pyx_vtabptr_8thriftrw_4wire_5value_I32Value;


/* "thriftrw/wire/value.pxd":85
 * 
 * 
 * cdef class I64Value(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int64_t value
 * 
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_I64Value {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_I64Value *__pyx_vtabptr_8thriftrw_4wire_5value_I64Value;


/* "thriftrw/wire/value.pxd":89
 * 
 * 
 * cdef class BinaryValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly bytes value
 *     # TODO change to char* once BinaryProtocol knows how to write that.
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_BinaryValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_BinaryValue *__pyx_vtabptr_8thriftrw_4wire_5value_BinaryValue;


/* "thriftrw/wire/value.pxd":100
 * 
 * 
 * cdef class StructValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly list fields
 *     cdef readonly dict _index
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_StructValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_StructValue *__pyx_vtabptr_8thriftrw_4wire_5value_StructValue;


/* "thriftrw/wire/value.pxd":110
 * 
 * 
 * cdef class MapValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t key_ttype
 *     cdef readonly int8_t value_ttype
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_MapValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_MapValue *__pyx_vtabptr_8thriftrw_4wire_5value_MapValue;


/* "thriftrw/wire/value.pxd":116
 * 
 * 
 * cdef class SetValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t value_ttype
 *     cdef readonly list values
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_SetValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_SetValue *__pyx_vtabptr_8thriftrw_4wire_5value_SetValue;


/* "thriftrw/wire/value.pxd":121
 * 
 * 
 * cdef class ListValue(Value):             # <<<<<<<<<<<<<<
 *     cdef readonly int8_t value_ttype
 *     cdef readonly list values
 */

struct __pyx_vtabstruct_8thriftrw_4wire_5value_ListValue {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_4wire_5value_ListValue *__pyx_vtabptr_8thriftrw_4wire_5value_ListValue;


/* "thriftrw/_buffer.pxd":24
 * 
 * 
 * cdef class ReadBuffer(object):             # <<<<<<<<<<<<<<
 *     cdef char* data
 *     cdef int offset, length
 */

struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer {
  void (*read)(struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *, char *, int, int __pyx_skip_dispatch);
  PyObject *(*take)(struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *, int, int __pyx_skip_dispatch);
  void (*skip)(struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *, int, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *__pyx_vtabptr_8thriftrw_7_buffer_ReadBuffer;


/* "thriftrw/_buffer.pxd":39
 * 
 * 
 * cdef class WriteBuffer(object):             # <<<<<<<<<<<<<<
 *     cdef char* data
 *     cdef readonly int length, capacity
 */

struct __pyx_vtabstruct_8thriftrw_7_buffer_WriteBuffer {
  void (*clear)(struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *, int __pyx_skip_dispatch);
  void (*write_bytes)(struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *, PyObject *, int __pyx_skip_dispatch);
  void (*write)(struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *, char *, int);
  void (*ensure_capacity)(struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *, int);
};
static struct __pyx_vtabstruct_8thriftrw_7_buffer_WriteBuffer *__pyx_vtabptr_8thriftrw_7_buffer_WriteBuffer;


/* "core.pxd":63
 * 
 * 
 * cdef class ProtocolWriter(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_value(self, Value value) except *
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolWriter {
  void (*write_value)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_obj_8thriftrw_4wire_5value_Value *);
  void (*write_bool)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int);
  void (*write_byte)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int8_t);
  void (*write_double)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, double);
  void (*write_i16)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int16_t);
  void (*write_i32)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int32_t);
  void (*write_i64)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int64_t);
  void (*write_binary)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, char *, int32_t);
  void (*write_struct_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *);
  void (*write_field_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader);
  void (*write_field_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *);
  void (*write_struct_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *);
  void (*write_map_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_MapHeader);
  void (*write_map_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *);
  void (*write_set_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_SetHeader);
  void (*write_set_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *);
  void (*write_list_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_ListHeader);
  void (*write_list_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *);
  void (*write_message_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *);
  void (*write_message_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *);
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolWriter *__pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolWriter;


/* "core.pxd":101
 * 
 * 
 * cdef class ProtocolReader:             # <<<<<<<<<<<<<<
 * 
 *     # Skip
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolReader {
  void (*skip)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *, int);
  void (*skip_binary)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*skip_map)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*skip_list)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*skip_set)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*skip_struct)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  int (*read_bool)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  int8_t (*read_byte)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  double (*read_double)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  int16_t (*read_i16)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  int32_t (*read_i32)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  int64_t (*read_i64)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  PyObject *(*read_binary)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*read_struct_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader (*read_field_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*read_field_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*read_struct_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  struct __pyx_t_8thriftrw_8protocol_4core_MapHeader (*read_map_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*read_map_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  struct __pyx_t_8thriftrw_8protocol_4core_SetHeader (*read_set_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*read_set_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  struct __pyx_t_8thriftrw_8protocol_4core_ListHeader (*read_list_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*read_list_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *(*read_message_begin)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
  void (*read_message_end)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *);
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolReader *__pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolReader;


/* "core.pxd":146
 * 
 * 
 * cdef class Protocol(object):             # <<<<<<<<<<<<<<
 *     cpdef ProtocolWriter writer(self, WriteBuffer buff)
 * 
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_4core_Protocol {
  struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *(*writer)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *, int __pyx_skip_dispatch);
  struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *(*reader)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *, int __pyx_skip_dispatch);
  PyObject *(*serialize_value)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, struct __pyx_obj_8thriftrw_4wire_5value_Value *, int __pyx_skip_dispatch);
  struct __pyx_obj_8thriftrw_4wire_5value_Value *(*deserialize_value)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, int, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*serialize_message)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, struct __pyx_obj_8thriftrw_4wire_7message_Message *, int __pyx_skip_dispatch);
  struct __pyx_obj_8thriftrw_4wire_7message_Message *(*deserialize_message)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_4core_Protocol *__pyx_vtabptr_8thriftrw_8protocol_4core_Protocol;


/* "core.pxd":160
 * 
 * 
 * cdef class _ValueWriter(ValueVisitor):             # <<<<<<<<<<<<<<
 *     cdef ProtocolWriter writer
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_4core__ValueWriter {
  struct __pyx_vtabstruct_8thriftrw_4wire_5value_ValueVisitor __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_4core__ValueWriter *__pyx_vtabptr_8thriftrw_8protocol_4core__ValueWriter;


/* "thriftrw/protocol/binary.pyx":362
 * 
 * 
 * cdef class BinaryProtocol(Protocol):             # <<<<<<<<<<<<<<
 *     """Implements the Thrift binary protocol."""
 * 
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocol {
  struct __pyx_vtabstruct_8thriftrw_8protocol_4core_Protocol __pyx_base;
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocol;


/* "thriftrw/protocol/binary.pyx":91
 * 
 * 
 * cdef class BinaryProtocolReader(ProtocolReader):             # <<<<<<<<<<<<<<
 *     """Parser for the binary protocol."""
 * 
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader {
  struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolReader __pyx_base;
  void (*_read)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *, char *, int);
  int8_t (*_byte)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *);
  int16_t (*_i16)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *);
  int32_t (*_i32)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *);
  int64_t (*_i64)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *);
  double (*_double)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *);
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolReader;


/* "thriftrw/protocol/binary.pyx":277
 * 
 * 
 * cdef class BinaryProtocolWriter(ProtocolWriter):             # <<<<<<<<<<<<<<
 *     """Serializes values using the Thrift Binary protocol."""
 * 
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter {
  struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolWriter __pyx_base;
  void (*_write)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *, char *, int);
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolWriter;


/* "thriftrw/protocol/binary.pyx":383
 * 
 * 
 * cdef class _OldBinaryProtocolReader(object):             # <<<<<<<<<<<<<<
 *     """Old version of BinaryProtocolReader that returned Value representations.
 * 
 */

struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader {
  PyObject *(*_reader)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *, int8_t);
  PyObject *(*read)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *, int8_t, int __pyx_skip_dispatch);
  void (*_read)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *, char *, int);
  int8_t (*_byte)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  int16_t (*_i16)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  int32_t (*_i32)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  int64_t (*_i64)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  double (*_double)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_7message_Message *(*read_message)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *(*read_bool)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *(*read_byte)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *(*read_double)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_I16Value *(*read_i16)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_I32Value *(*read_i32)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_I64Value *(*read_i64)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *(*read_binary)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_StructValue *(*read_struct)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_MapValue *(*read_map)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_SetValue *(*read_set)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
  struct __pyx_obj_8thriftrw_4wire_5value_ListValue *(*read_list)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *);
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_vtabptr_8thriftrw_8protocol_6binary__OldBinaryProtocolReader;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* IncludeStringH.proto */
#include <string.h>

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int8_t(int8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int16_t(int16_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int8_t __Pyx_PyInt_As_int8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int16_t __Pyx_PyInt_As_int16_t(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, int __pyx_v_typ); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_struct(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_map(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_list(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_set(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__read(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, char *__pyx_v_data, int __pyx_v_count); /* proto*/
static int8_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__byte(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int16_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i16(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int32_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i32(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int64_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i64(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static double __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__double(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_bool(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int8_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_byte(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static double __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_double(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int16_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i16(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int32_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i32(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static int64_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i64(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_binary(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_binary(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_field_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_map_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_set_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_list_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_message_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter__write(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, char *__pyx_v_data, int __pyx_v_length); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_bool(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int __pyx_v_value); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_byte(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int8_t __pyx_v_value); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_double(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, double __pyx_v_value); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i16(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int16_t __pyx_v_value); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i32(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int32_t __pyx_v_value); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i64(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int64_t __pyx_v_value); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_binary(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, char *__pyx_v_value, int32_t __pyx_v_length); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_field_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_v_header); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_struct_end(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_map_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_v_header); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_set_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_v_header); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_list_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_v_header); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_message_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *__pyx_v_message); /* proto*/
static struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_reader(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_buff, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_writer(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *__pyx_v_buff, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_7message_Message *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_message(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, PyObject *__pyx_v_s, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_Value *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_value(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, int __pyx_v_typ, PyObject *__pyx_v_s, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__reader(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, int8_t __pyx_v_typ); /* proto*/
static PyObject *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, int8_t __pyx_v_typ, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__read(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, char *__pyx_v_data, int __pyx_v_count); /* proto*/
static int8_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__byte(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static int16_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i16(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static int32_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i32(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static int64_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i64(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static double __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__double(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_7message_Message *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_message(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_bool(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_byte(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_double(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_I16Value *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i16(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_I32Value *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i32(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_I64Value *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i64(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_binary(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_StructValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_struct(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_MapValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_map(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_SetValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_set(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_ListValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_list(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'thriftrw.wire' */

/* Module declarations from 'thriftrw.wire.ttype' */

/* Module declarations from 'thriftrw.wire.value' */
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_ValueVisitor = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value__ValueVisitorArgs = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_Value = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_BoolValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_ByteValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_DoubleValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_I16Value = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_I32Value = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_I64Value = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_BinaryValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_FieldValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_StructValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_MapItem = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_MapValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_SetValue = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_5value_ListValue = 0;

/* Module declarations from 'thriftrw.wire.message' */
static PyTypeObject *__pyx_ptype_8thriftrw_4wire_7message_Message = 0;

/* Module declarations from 'thriftrw._buffer' */
static PyTypeObject *__pyx_ptype_8thriftrw_7_buffer_ReadBuffer = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_7_buffer_WriteBuffer = 0;

/* Module declarations from 'thriftrw.protocol.core' */
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_4core_MessageHeader = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_4core_Protocol = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_4core__ValueWriter = 0;

/* Module declarations from 'thriftrw.protocol._endian' */

/* Module declarations from 'thriftrw.protocol.binary' */
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocol = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocolReader = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocolWriter = 0;
static PyTypeObject *__pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py = 0;
static PyObject *__pyx_v_8thriftrw_8protocol_6binary_STRUCT_END_HEADER = 0;
static int8_t __pyx_v_8thriftrw_8protocol_6binary_STRUCT_END;
static int32_t __pyx_v_8thriftrw_8protocol_6binary_VERSION;
static int32_t __pyx_v_8thriftrw_8protocol_6binary_VERSION_MASK;
static int32_t __pyx_v_8thriftrw_8protocol_6binary_TYPE_MASK;
static PyObject *__pyx_f_8thriftrw_8protocol_6binary___pyx_unpickle_BinaryProtocol__set_state(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *, PyObject *); /*proto*/
static struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader(PyObject *); /*proto*/
static PyObject *__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_I16Value *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_I32Value *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_I64Value *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_StructValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_MapValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_SetValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
static PyObject *__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_ListValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)); /*proto*/
#define __Pyx_MODULE_NAME "thriftrw.protocol.binary"
extern int __pyx_module_is_main_thriftrw__protocol__binary;
int __pyx_module_is_main_thriftrw__protocol__binary = 0;

/* Implementation of 'thriftrw.protocol.binary' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_s[] = "s";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_typ[] = "typ";
static const char __pyx_k_body[] = "body";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_wrap[] = "wrap";
static const char __pyx_k_ktype[] = "ktype";
static const char __pyx_k_pairs[] = "pairs";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_seqid[] = "seqid";
static const char __pyx_k_ttype[] = "ttype";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_vtype[] = "vtype";
static const char __pyx_k_errors[] = "errors";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reader[] = "reader";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_writer[] = "writer";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_key_ttype[] = "key_ttype";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_cfunc_to_py[] = "cfunc.to_py";
static const char __pyx_k_value_ttype[] = "value_ttype";
static const char __pyx_k_message_type[] = "message_type";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_BinaryProtocol[] = "BinaryProtocol";
static const char __pyx_k_EndOfInputError[] = "EndOfInputError";
static const char __pyx_k_Unknown_TType_r[] = "Unknown TType \"%r\"";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_deserialize_value[] = "deserialize_value";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_ThriftProtocolError[] = "ThriftProtocolError";
static const char __pyx_k_deserialize_message[] = "deserialize_message";
static const char __pyx_k_Unsupported_version_r[] = "Unsupported version \"%r\"";
static const char __pyx_k_thriftrw_protocol_binary[] = "thriftrw.protocol.binary";
static const char __pyx_k_pyx_unpickle_BinaryProtocol[] = "__pyx_unpickle_BinaryProtocol";
static const char __pyx_k_Pyx_CFunc_BinaryValue______Old[] = "__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_BoolValue______OldBi[] = "__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_ByteValue______OldBi[] = "__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_DoubleValue______Old[] = "__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_I16Value______OldBin[] = "__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_I32Value______OldBin[] = "__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_I64Value______OldBin[] = "__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_ListValue______OldBi[] = "__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_MapValue______OldBin[] = "__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_SetValue______OldBin[] = "__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Pyx_CFunc_StructValue______Old[] = "__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py.<locals>.wrap";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static const char __pyx_k_No_value_specified_for_struct_at[] = "No value specified for struct attribute 'type'";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_No_value_specified_for_struct_at_2[] = "No value specified for struct attribute 'id'";
static PyObject *__pyx_n_u_BinaryProtocol;
static PyObject *__pyx_n_s_EndOfInputError;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at;
static PyObject *__pyx_kp_s_No_value_specified_for_struct_at_2;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Pyx_CFunc_BinaryValue______Old;
static PyObject *__pyx_n_s_Pyx_CFunc_BoolValue______OldBi;
static PyObject *__pyx_n_s_Pyx_CFunc_ByteValue______OldBi;
static PyObject *__pyx_n_s_Pyx_CFunc_DoubleValue______Old;
static PyObject *__pyx_n_s_Pyx_CFunc_I16Value______OldBin;
static PyObject *__pyx_n_s_Pyx_CFunc_I32Value______OldBin;
static PyObject *__pyx_n_s_Pyx_CFunc_I64Value______OldBin;
static PyObject *__pyx_n_s_Pyx_CFunc_ListValue______OldBi;
static PyObject *__pyx_n_s_Pyx_CFunc_MapValue______OldBin;
static PyObject *__pyx_n_s_Pyx_CFunc_SetValue______OldBin;
static PyObject *__pyx_n_s_Pyx_CFunc_StructValue______Old;
static PyObject *__pyx_n_s_ThriftProtocolError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_u_Unknown_TType_r;
static PyObject *__pyx_kp_u_Unsupported_version_r;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_body;
static PyObject *__pyx_n_s_cfunc_to_py;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_deserialize_message;
static PyObject *__pyx_n_s_deserialize_value;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_errors;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_key_ttype;
static PyObject *__pyx_n_s_ktype;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_message_type;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_pairs;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_BinaryProtocol;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_reader;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_seqid;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_thriftrw_protocol_binary;
static PyObject *__pyx_n_s_ttype;
static PyObject *__pyx_n_s_typ;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_value_ttype;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_vtype;
static PyObject *__pyx_n_s_wrap;
static PyObject *__pyx_n_s_writer;
static int __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader___cinit__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_reader); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter___cinit__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *__pyx_v_writer); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_reader(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_buff); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_2writer(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *__pyx_v_buff); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_4deserialize_message(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_6deserialize_value(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, int __pyx_v_typ, PyObject *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_8__reduce_cython__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_10__setstate_cython__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader___cinit__(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_reader); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_2read(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, int8_t __pyx_v_typ); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_8thriftrw_8protocol_6binary___pyx_unpickle_BinaryProtocol(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self); /* proto */
static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocol(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocolReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocolWriter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary__OldBinaryProtocolReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_codeobj__10;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
/* Late includes */

/* "thriftrw/protocol/binary.pyx":94
 *     """Parser for the binary protocol."""
 * 
 *     def __cinit__(self, ReadBuffer reader):             # <<<<<<<<<<<<<<
 *         """Initialize the reader.
 * 
 */

/* Python wrapper */
static int __pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_reader = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_reader,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reader)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 94, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_reader = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 94, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_reader), __pyx_ptype_8thriftrw_7_buffer_ReadBuffer, 1, "reader", 0))) __PYX_ERR(0, 94, __pyx_L1_error)
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader___cinit__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self), __pyx_v_reader);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader___cinit__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_reader) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "thriftrw/protocol/binary.pyx":102
 *             ``num`` is negative.
 *         """
 *         self.reader = reader             # <<<<<<<<<<<<<<
 * 
 *     cdef void skip(self, int typ) except *:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_reader));
  __Pyx_GOTREF(__pyx_v_self->reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->reader));
  __pyx_v_self->reader = __pyx_v_reader;

  /* "thriftrw/protocol/binary.pyx":94
 *     """Parser for the binary protocol."""
 * 
 *     def __cinit__(self, ReadBuffer reader):             # <<<<<<<<<<<<<<
 *         """Initialize the reader.
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":104
 *         self.reader = reader
 * 
 *     cdef void skip(self, int typ) except *:             # <<<<<<<<<<<<<<
 *         if typ == ttype.BOOL:
 *             self.reader.skip(1)
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, int __pyx_v_typ) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("skip", 0);

  /* "thriftrw/protocol/binary.pyx":105
 * 
 *     cdef void skip(self, int typ) except *:
 *         if typ == ttype.BOOL:             # <<<<<<<<<<<<<<
 *             self.reader.skip(1)
 *         elif typ == ttype.BYTE:
 */
  switch (__pyx_v_typ) {
    case __pyx_e_8thriftrw_4wire_5ttype_BOOL:

    /* "thriftrw/protocol/binary.pyx":106
 *     cdef void skip(self, int typ) except *:
 *         if typ == ttype.BOOL:
 *             self.reader.skip(1)             # <<<<<<<<<<<<<<
 *         elif typ == ttype.BYTE:
 *             self.reader.skip(1)
 */
    ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->skip(__pyx_v_self->reader, 1, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 106, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":105
 * 
 *     cdef void skip(self, int typ) except *:
 *         if typ == ttype.BOOL:             # <<<<<<<<<<<<<<
 *             self.reader.skip(1)
 *         elif typ == ttype.BYTE:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":107
 *         if typ == ttype.BOOL:
 *             self.reader.skip(1)
 *         elif typ == ttype.BYTE:             # <<<<<<<<<<<<<<
 *             self.reader.skip(1)
 *         elif typ == ttype.DOUBLE:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_BYTE:

    /* "thriftrw/protocol/binary.pyx":108
 *             self.reader.skip(1)
 *         elif typ == ttype.BYTE:
 *             self.reader.skip(1)             # <<<<<<<<<<<<<<
 *         elif typ == ttype.DOUBLE:
 *             self.reader.skip(8)
 */
    ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->skip(__pyx_v_self->reader, 1, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 108, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":107
 *         if typ == ttype.BOOL:
 *             self.reader.skip(1)
 *         elif typ == ttype.BYTE:             # <<<<<<<<<<<<<<
 *             self.reader.skip(1)
 *         elif typ == ttype.DOUBLE:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":109
 *         elif typ == ttype.BYTE:
 *             self.reader.skip(1)
 *         elif typ == ttype.DOUBLE:             # <<<<<<<<<<<<<<
 *             self.reader.skip(8)
 *         elif typ == ttype.I16:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_DOUBLE:

    /* "thriftrw/protocol/binary.pyx":110
 *             self.reader.skip(1)
 *         elif typ == ttype.DOUBLE:
 *             self.reader.skip(8)             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I16:
 *             self.reader.skip(2)
 */
    ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->skip(__pyx_v_self->reader, 8, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 110, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":109
 *         elif typ == ttype.BYTE:
 *             self.reader.skip(1)
 *         elif typ == ttype.DOUBLE:             # <<<<<<<<<<<<<<
 *             self.reader.skip(8)
 *         elif typ == ttype.I16:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":111
 *         elif typ == ttype.DOUBLE:
 *             self.reader.skip(8)
 *         elif typ == ttype.I16:             # <<<<<<<<<<<<<<
 *             self.reader.skip(2)
 *         elif typ == ttype.I32:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I16:

    /* "thriftrw/protocol/binary.pyx":112
 *             self.reader.skip(8)
 *         elif typ == ttype.I16:
 *             self.reader.skip(2)             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I32:
 *             self.reader.skip(4)
 */
    ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->skip(__pyx_v_self->reader, 2, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 112, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":111
 *         elif typ == ttype.DOUBLE:
 *             self.reader.skip(8)
 *         elif typ == ttype.I16:             # <<<<<<<<<<<<<<
 *             self.reader.skip(2)
 *         elif typ == ttype.I32:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":113
 *         elif typ == ttype.I16:
 *             self.reader.skip(2)
 *         elif typ == ttype.I32:             # <<<<<<<<<<<<<<
 *             self.reader.skip(4)
 *         elif typ == ttype.I64:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I32:

    /* "thriftrw/protocol/binary.pyx":114
 *             self.reader.skip(2)
 *         elif typ == ttype.I32:
 *             self.reader.skip(4)             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I64:
 *             self.reader.skip(8)
 */
    ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->skip(__pyx_v_self->reader, 4, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 114, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":113
 *         elif typ == ttype.I16:
 *             self.reader.skip(2)
 *         elif typ == ttype.I32:             # <<<<<<<<<<<<<<
 *             self.reader.skip(4)
 *         elif typ == ttype.I64:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":115
 *         elif typ == ttype.I32:
 *             self.reader.skip(4)
 *         elif typ == ttype.I64:             # <<<<<<<<<<<<<<
 *             self.reader.skip(8)
 *         elif typ == ttype.BINARY:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I64:

    /* "thriftrw/protocol/binary.pyx":116
 *             self.reader.skip(4)
 *         elif typ == ttype.I64:
 *             self.reader.skip(8)             # <<<<<<<<<<<<<<
 *         elif typ == ttype.BINARY:
 *             self.skip_binary()
 */
    ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->skip(__pyx_v_self->reader, 8, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 116, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":115
 *         elif typ == ttype.I32:
 *             self.reader.skip(4)
 *         elif typ == ttype.I64:             # <<<<<<<<<<<<<<
 *             self.reader.skip(8)
 *         elif typ == ttype.BINARY:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":117
 *         elif typ == ttype.I64:
 *             self.reader.skip(8)
 *         elif typ == ttype.BINARY:             # <<<<<<<<<<<<<<
 *             self.skip_binary()
 *         elif typ == ttype.STRUCT:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_BINARY:

    /* "thriftrw/protocol/binary.pyx":118
 *             self.reader.skip(8)
 *         elif typ == ttype.BINARY:
 *             self.skip_binary()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.STRUCT:
 *             self.skip_struct()
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip_binary(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 118, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":117
 *         elif typ == ttype.I64:
 *             self.reader.skip(8)
 *         elif typ == ttype.BINARY:             # <<<<<<<<<<<<<<
 *             self.skip_binary()
 *         elif typ == ttype.STRUCT:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":119
 *         elif typ == ttype.BINARY:
 *             self.skip_binary()
 *         elif typ == ttype.STRUCT:             # <<<<<<<<<<<<<<
 *             self.skip_struct()
 *         elif typ == ttype.MAP:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_STRUCT:

    /* "thriftrw/protocol/binary.pyx":120
 *             self.skip_binary()
 *         elif typ == ttype.STRUCT:
 *             self.skip_struct()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.MAP:
 *             self.skip_map()
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip_struct(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 120, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":119
 *         elif typ == ttype.BINARY:
 *             self.skip_binary()
 *         elif typ == ttype.STRUCT:             # <<<<<<<<<<<<<<
 *             self.skip_struct()
 *         elif typ == ttype.MAP:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":121
 *         elif typ == ttype.STRUCT:
 *             self.skip_struct()
 *         elif typ == ttype.MAP:             # <<<<<<<<<<<<<<
 *             self.skip_map()
 *         elif typ == ttype.SET:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_MAP:

    /* "thriftrw/protocol/binary.pyx":122
 *             self.skip_struct()
 *         elif typ == ttype.MAP:
 *             self.skip_map()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.SET:
 *             self.skip_set()
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip_map(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 122, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":121
 *         elif typ == ttype.STRUCT:
 *             self.skip_struct()
 *         elif typ == ttype.MAP:             # <<<<<<<<<<<<<<
 *             self.skip_map()
 *         elif typ == ttype.SET:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":123
 *         elif typ == ttype.MAP:
 *             self.skip_map()
 *         elif typ == ttype.SET:             # <<<<<<<<<<<<<<
 *             self.skip_set()
 *         elif typ == ttype.LIST:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_SET:

    /* "thriftrw/protocol/binary.pyx":124
 *             self.skip_map()
 *         elif typ == ttype.SET:
 *             self.skip_set()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.LIST:
 *             self.skip_list()
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip_set(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":123
 *         elif typ == ttype.MAP:
 *             self.skip_map()
 *         elif typ == ttype.SET:             # <<<<<<<<<<<<<<
 *             self.skip_set()
 *         elif typ == ttype.LIST:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":125
 *         elif typ == ttype.SET:
 *             self.skip_set()
 *         elif typ == ttype.LIST:             # <<<<<<<<<<<<<<
 *             self.skip_list()
 * 
 */
    case __pyx_e_8thriftrw_4wire_5ttype_LIST:

    /* "thriftrw/protocol/binary.pyx":126
 *             self.skip_set()
 *         elif typ == ttype.LIST:
 *             self.skip_list()             # <<<<<<<<<<<<<<
 * 
 *     cdef void skip_struct(self) except *:
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip_list(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 126, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":125
 *         elif typ == ttype.SET:
 *             self.skip_set()
 *         elif typ == ttype.LIST:             # <<<<<<<<<<<<<<
 *             self.skip_list()
 * 
 */
    break;
    default: break;
  }

  /* "thriftrw/protocol/binary.pyx":104
 *         self.reader = reader
 * 
 *     cdef void skip(self, int typ) except *:             # <<<<<<<<<<<<<<
 *         if typ == ttype.BOOL:
 *             self.reader.skip(1)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.skip", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":128
 *             self.skip_list()
 * 
 *     cdef void skip_struct(self) except *:             # <<<<<<<<<<<<<<
 *         header = self.read_field_begin()
 *         while header.type != -1:
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_struct(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_v_header;
  __Pyx_RefNannyDeclarations
  struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("skip_struct", 0);

  /* "thriftrw/protocol/binary.pyx":129
 * 
 *     cdef void skip_struct(self) except *:
 *         header = self.read_field_begin()             # <<<<<<<<<<<<<<
 *         while header.type != -1:
 *             self.skip(header.type)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.read_field_begin(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 129, __pyx_L1_error)
  __pyx_v_header = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":130
 *     cdef void skip_struct(self) except *:
 *         header = self.read_field_begin()
 *         while header.type != -1:             # <<<<<<<<<<<<<<
 *             self.skip(header.type)
 *             header = self.read_field_begin()
 */
  while (1) {
    __pyx_t_2 = ((__pyx_v_header.type != -1L) != 0);
    if (!__pyx_t_2) break;

    /* "thriftrw/protocol/binary.pyx":131
 *         header = self.read_field_begin()
 *         while header.type != -1:
 *             self.skip(header.type)             # <<<<<<<<<<<<<<
 *             header = self.read_field_begin()
 * 
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self), __pyx_v_header.type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 131, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":132
 *         while header.type != -1:
 *             self.skip(header.type)
 *             header = self.read_field_begin()             # <<<<<<<<<<<<<<
 * 
 *     cdef void skip_map(self) except *:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.read_field_begin(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 132, __pyx_L1_error)
    __pyx_v_header = __pyx_t_1;
  }

  /* "thriftrw/protocol/binary.pyx":128
 *             self.skip_list()
 * 
 *     cdef void skip_struct(self) except *:             # <<<<<<<<<<<<<<
 *         header = self.read_field_begin()
 *         while header.type != -1:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.skip_struct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":134
 *             header = self.read_field_begin()
 * 
 *     cdef void skip_map(self) except *:             # <<<<<<<<<<<<<<
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_map(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_v_header;
  CYTHON_UNUSED int32_t __pyx_v__;
  __Pyx_RefNannyDeclarations
  struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_t_1;
  int32_t __pyx_t_2;
  int32_t __pyx_t_3;
  int32_t __pyx_t_4;
  __Pyx_RefNannySetupContext("skip_map", 0);

  /* "thriftrw/protocol/binary.pyx":137
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 *         cdef MapHeader header = self.read_map_begin()             # <<<<<<<<<<<<<<
 *         for _ in range(header.size):
 *             self.skip(header.ktype)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.read_map_begin(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 137, __pyx_L1_error)
  __pyx_v_header = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":138
 *         # size for fixed width types
 *         cdef MapHeader header = self.read_map_begin()
 *         for _ in range(header.size):             # <<<<<<<<<<<<<<
 *             self.skip(header.ktype)
 *             self.skip(header.vtype)
 */
  __pyx_t_2 = __pyx_v_header.size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v__ = __pyx_t_4;

    /* "thriftrw/protocol/binary.pyx":139
 *         cdef MapHeader header = self.read_map_begin()
 *         for _ in range(header.size):
 *             self.skip(header.ktype)             # <<<<<<<<<<<<<<
 *             self.skip(header.vtype)
 * 
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self), __pyx_v_header.ktype); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 139, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":140
 *         for _ in range(header.size):
 *             self.skip(header.ktype)
 *             self.skip(header.vtype)             # <<<<<<<<<<<<<<
 * 
 *     cdef void skip_list(self) except *:
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self), __pyx_v_header.vtype); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L1_error)
  }

  /* "thriftrw/protocol/binary.pyx":134
 *             header = self.read_field_begin()
 * 
 *     cdef void skip_map(self) except *:             # <<<<<<<<<<<<<<
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.skip_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":142
 *             self.skip(header.vtype)
 * 
 *     cdef void skip_list(self) except *:             # <<<<<<<<<<<<<<
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_list(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_v_header;
  CYTHON_UNUSED int32_t __pyx_v__;
  __Pyx_RefNannyDeclarations
  struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_t_1;
  int32_t __pyx_t_2;
  int32_t __pyx_t_3;
  int32_t __pyx_t_4;
  __Pyx_RefNannySetupContext("skip_list", 0);

  /* "thriftrw/protocol/binary.pyx":145
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 *         cdef ListHeader header = self.read_list_begin()             # <<<<<<<<<<<<<<
 *         for _ in range(header.size):
 *             self.skip(header.type)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.read_list_begin(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L1_error)
  __pyx_v_header = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":146
 *         # size for fixed width types
 *         cdef ListHeader header = self.read_list_begin()
 *         for _ in range(header.size):             # <<<<<<<<<<<<<<
 *             self.skip(header.type)
 * 
 */
  __pyx_t_2 = __pyx_v_header.size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v__ = __pyx_t_4;

    /* "thriftrw/protocol/binary.pyx":147
 *         cdef ListHeader header = self.read_list_begin()
 *         for _ in range(header.size):
 *             self.skip(header.type)             # <<<<<<<<<<<<<<
 * 
 *     cdef void skip_set(self) except *:
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self), __pyx_v_header.type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 147, __pyx_L1_error)
  }

  /* "thriftrw/protocol/binary.pyx":142
 *             self.skip(header.vtype)
 * 
 *     cdef void skip_list(self) except *:             # <<<<<<<<<<<<<<
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.skip_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":149
 *             self.skip(header.type)
 * 
 *     cdef void skip_set(self) except *:             # <<<<<<<<<<<<<<
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_set(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_v_header;
  CYTHON_UNUSED int32_t __pyx_v__;
  __Pyx_RefNannyDeclarations
  struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_t_1;
  int32_t __pyx_t_2;
  int32_t __pyx_t_3;
  int32_t __pyx_t_4;
  __Pyx_RefNannySetupContext("skip_set", 0);

  /* "thriftrw/protocol/binary.pyx":152
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 *         cdef SetHeader header = self.read_set_begin()             # <<<<<<<<<<<<<<
 *         for _ in range(header.size):
 *             self.skip(header.type)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.read_set_begin(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_v_header = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":153
 *         # size for fixed width types
 *         cdef SetHeader header = self.read_set_begin()
 *         for _ in range(header.size):             # <<<<<<<<<<<<<<
 *             self.skip(header.type)
 * 
 */
  __pyx_t_2 = __pyx_v_header.size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v__ = __pyx_t_4;

    /* "thriftrw/protocol/binary.pyx":154
 *         cdef SetHeader header = self.read_set_begin()
 *         for _ in range(header.size):
 *             self.skip(header.type)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _read(self, char* data, int count) except *:
 */
    ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.skip(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_v_self), __pyx_v_header.type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L1_error)
  }

  /* "thriftrw/protocol/binary.pyx":149
 *             self.skip(header.type)
 * 
 *     cdef void skip_set(self) except *:             # <<<<<<<<<<<<<<
 *         # TODO: This could be optimized to precompute the expected
 *         # size for fixed width types
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.skip_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":156
 *             self.skip(header.type)
 * 
 *     cdef void _read(self, char* data, int count) except *:             # <<<<<<<<<<<<<<
 *         self.reader.read(data, count)
 * 
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__read(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, char *__pyx_v_data, int __pyx_v_count) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read", 0);

  /* "thriftrw/protocol/binary.pyx":157
 * 
 *     cdef void _read(self, char* data, int count) except *:
 *         self.reader.read(data, count)             # <<<<<<<<<<<<<<
 * 
 *     cdef int8_t _byte(self) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->read(__pyx_v_self->reader, __pyx_v_data, __pyx_v_count, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":156
 *             self.skip(header.type)
 * 
 *     cdef void _read(self, char* data, int count) except *:             # <<<<<<<<<<<<<<
 *         self.reader.read(data, count)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader._read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":159
 *         self.reader.read(data, count)
 * 
 *     cdef int8_t _byte(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char c
 *         self._read(&c, 1)
 */

static int8_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__byte(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_c;
  int8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_byte", 0);

  /* "thriftrw/protocol/binary.pyx":161
 *     cdef int8_t _byte(self) except *:
 *         cdef char c
 *         self._read(&c, 1)             # <<<<<<<<<<<<<<
 *         return <int8_t>c
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_read(__pyx_v_self, (&__pyx_v_c), 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 161, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":162
 *         cdef char c
 *         self._read(&c, 1)
 *         return <int8_t>c             # <<<<<<<<<<<<<<
 * 
 *     cdef int16_t _i16(self) except *:
 */
  __pyx_r = ((int8_t)__pyx_v_c);
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":159
 *         self.reader.read(data, count)
 * 
 *     cdef int8_t _byte(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char c
 *         self._read(&c, 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader._byte", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":164
 *         return <int8_t>c
 * 
 *     cdef int16_t _i16(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[2]
 *         self._read(data, 2)
 */

static int16_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i16(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_data[2];
  int16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_i16", 0);

  /* "thriftrw/protocol/binary.pyx":166
 *     cdef int16_t _i16(self) except *:
 *         cdef char data[2]
 *         self._read(data, 2)             # <<<<<<<<<<<<<<
 *         return be16toh((<int16_t*>data)[0])
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_read(__pyx_v_self, __pyx_v_data, 2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":167
 *         cdef char data[2]
 *         self._read(data, 2)
 *         return be16toh((<int16_t*>data)[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef int32_t _i32(self) except *:
 */
  __pyx_r = be16toh((((int16_t *)__pyx_v_data)[0]));
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":164
 *         return <int8_t>c
 * 
 *     cdef int16_t _i16(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[2]
 *         self._read(data, 2)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader._i16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":169
 *         return be16toh((<int16_t*>data)[0])
 * 
 *     cdef int32_t _i32(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[4]
 *         self._read(data, 4)
 */

static int32_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i32(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_data[4];
  int32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_i32", 0);

  /* "thriftrw/protocol/binary.pyx":171
 *     cdef int32_t _i32(self) except *:
 *         cdef char data[4]
 *         self._read(data, 4)             # <<<<<<<<<<<<<<
 *         return be32toh((<int32_t*>data)[0])
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_read(__pyx_v_self, __pyx_v_data, 4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":172
 *         cdef char data[4]
 *         self._read(data, 4)
 *         return be32toh((<int32_t*>data)[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef int64_t _i64(self) except *:
 */
  __pyx_r = be32toh((((int32_t *)__pyx_v_data)[0]));
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":169
 *         return be16toh((<int16_t*>data)[0])
 * 
 *     cdef int32_t _i32(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[4]
 *         self._read(data, 4)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader._i32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":174
 *         return be32toh((<int32_t*>data)[0])
 * 
 *     cdef int64_t _i64(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[8]
 *         self._read(data, 8)
 */

static int64_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i64(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_data[8];
  int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_i64", 0);

  /* "thriftrw/protocol/binary.pyx":176
 *     cdef int64_t _i64(self) except *:
 *         cdef char data[8]
 *         self._read(data, 8)             # <<<<<<<<<<<<<<
 *         return be64toh((<int64_t*>data)[0])
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_read(__pyx_v_self, __pyx_v_data, 8); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 176, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":177
 *         cdef char data[8]
 *         self._read(data, 8)
 *         return be64toh((<int64_t*>data)[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef double _double(self) except *:
 */
  __pyx_r = be64toh((((int64_t *)__pyx_v_data)[0]));
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":174
 *         return be32toh((<int32_t*>data)[0])
 * 
 *     cdef int64_t _i64(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[8]
 *         self._read(data, 8)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader._i64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":179
 *         return be64toh((<int64_t*>data)[0])
 * 
 *     cdef double _double(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int64_t value = self._i64()
 *         return (<double*>(&value))[0]
 */

static double __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__double(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int64_t __pyx_v_value;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  __Pyx_RefNannySetupContext("_double", 0);

  /* "thriftrw/protocol/binary.pyx":180
 * 
 *     cdef double _double(self) except *:
 *         cdef int64_t value = self._i64()             # <<<<<<<<<<<<<<
 *         return (<double*>(&value))[0]
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i64(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 180, __pyx_L1_error)
  __pyx_v_value = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":181
 *     cdef double _double(self) except *:
 *         cdef int64_t value = self._i64()
 *         return (<double*>(&value))[0]             # <<<<<<<<<<<<<<
 * 
 *     cdef bint read_bool(self) except *:
 */
  __pyx_r = (((double *)(&__pyx_v_value))[0]);
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":179
 *         return be64toh((<int64_t*>data)[0])
 * 
 *     cdef double _double(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int64_t value = self._i64()
 *         return (<double*>(&value))[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader._double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":183
 *         return (<double*>(&value))[0]
 * 
 *     cdef bint read_bool(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a boolean."""
 *         return self._byte() == 1
 */

static int __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_bool(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  __Pyx_RefNannySetupContext("read_bool", 0);

  /* "thriftrw/protocol/binary.pyx":185
 *     cdef bint read_bool(self) except *:
 *         """Reads a boolean."""
 *         return self._byte() == 1             # <<<<<<<<<<<<<<
 * 
 *     cdef int8_t read_byte(self) except *:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 185, __pyx_L1_error)
  __pyx_r = (__pyx_t_1 == 1);
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":183
 *         return (<double*>(&value))[0]
 * 
 *     cdef bint read_bool(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a boolean."""
 *         return self._byte() == 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_bool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":187
 *         return self._byte() == 1
 * 
 *     cdef int8_t read_byte(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a byte."""
 *         return self._byte()
 */

static int8_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_byte(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  __Pyx_RefNannySetupContext("read_byte", 0);

  /* "thriftrw/protocol/binary.pyx":189
 *     cdef int8_t read_byte(self) except *:
 *         """Reads a byte."""
 *         return self._byte()             # <<<<<<<<<<<<<<
 * 
 *     cdef double read_double(self) except *:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":187
 *         return self._byte() == 1
 * 
 *     cdef int8_t read_byte(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a byte."""
 *         return self._byte()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_byte", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":191
 *         return self._byte()
 * 
 *     cdef double read_double(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a double."""
 *         return self._double()
 */

static double __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_double(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("read_double", 0);

  /* "thriftrw/protocol/binary.pyx":193
 *     cdef double read_double(self) except *:
 *         """Reads a double."""
 *         return self._double()             # <<<<<<<<<<<<<<
 * 
 *     cdef int16_t read_i16(self) except *:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_double(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 193, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":191
 *         return self._byte()
 * 
 *     cdef double read_double(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a double."""
 *         return self._double()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":195
 *         return self._double()
 * 
 *     cdef int16_t read_i16(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a 16-bit integer."""
 *         return self._i16()
 */

static int16_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i16(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int16_t __pyx_t_1;
  __Pyx_RefNannySetupContext("read_i16", 0);

  /* "thriftrw/protocol/binary.pyx":197
 *     cdef int16_t read_i16(self) except *:
 *         """Reads a 16-bit integer."""
 *         return self._i16()             # <<<<<<<<<<<<<<
 * 
 *     cdef int32_t read_i32(self) except *:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i16(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":195
 *         return self._double()
 * 
 *     cdef int16_t read_i16(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a 16-bit integer."""
 *         return self._i16()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_i16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":199
 *         return self._i16()
 * 
 *     cdef int32_t read_i32(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a 32-bit integer."""
 *         return self._i32()
 */

static int32_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i32(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("read_i32", 0);

  /* "thriftrw/protocol/binary.pyx":201
 *     cdef int32_t read_i32(self) except *:
 *         """Reads a 32-bit integer."""
 *         return self._i32()             # <<<<<<<<<<<<<<
 * 
 *     cdef int64_t read_i64(self) except *:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 201, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":199
 *         return self._i16()
 * 
 *     cdef int32_t read_i32(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a 32-bit integer."""
 *         return self._i32()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_i32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":203
 *         return self._i32()
 * 
 *     cdef int64_t read_i64(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a 64-bit integer."""
 *         return self._i64()
 */

static int64_t __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i64(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  __Pyx_RefNannySetupContext("read_i64", 0);

  /* "thriftrw/protocol/binary.pyx":205
 *     cdef int64_t read_i64(self) except *:
 *         """Reads a 64-bit integer."""
 *         return self._i64()             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes read_binary(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i64(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 205, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":203
 *         return self._i32()
 * 
 *     cdef int64_t read_i64(self) except *:             # <<<<<<<<<<<<<<
 *         """Reads a 64-bit integer."""
 *         return self._i64()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_i64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":207
 *         return self._i64()
 * 
 *     cdef bytes read_binary(self):             # <<<<<<<<<<<<<<
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()
 */

static PyObject *__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_binary(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int32_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("read_binary", 0);

  /* "thriftrw/protocol/binary.pyx":209
 *     cdef bytes read_binary(self):
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 *         return self.reader.take(length)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 209, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":210
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()
 *         return self.reader.take(length)             # <<<<<<<<<<<<<<
 * 
 *     cdef void skip_binary(self) except *:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->take(__pyx_v_self->reader, __pyx_v_length, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":207
 *         return self._i64()
 * 
 *     cdef bytes read_binary(self):             # <<<<<<<<<<<<<<
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_binary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":212
 *         return self.reader.take(length)
 * 
 *     cdef void skip_binary(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 *         self.reader.skip(length)
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_binary(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int32_t __pyx_v_length;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  __Pyx_RefNannySetupContext("skip_binary", 0);

  /* "thriftrw/protocol/binary.pyx":213
 * 
 *     cdef void skip_binary(self) except *:
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 *         self.reader.skip(length)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 213, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":214
 *     cdef void skip_binary(self) except *:
 *         cdef int32_t length = self._i32()
 *         self.reader.skip(length)             # <<<<<<<<<<<<<<
 * 
 *     cdef FieldHeader read_field_begin(self) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->skip(__pyx_v_self->reader, __pyx_v_length, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":212
 *         return self.reader.take(length)
 * 
 *     cdef void skip_binary(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 *         self.reader.skip(length)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.skip_binary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":216
 *         self.reader.skip(length)
 * 
 *     cdef FieldHeader read_field_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t field_type = self._byte()
 *         if field_type == STRUCT_END:
 */

static struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_field_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_field_type;
  int16_t __pyx_v_field_id;
  struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_r;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_t_3;
  int16_t __pyx_t_4;
  __Pyx_RefNannySetupContext("read_field_begin", 0);

  /* "thriftrw/protocol/binary.pyx":217
 * 
 *     cdef FieldHeader read_field_begin(self) except *:
 *         cdef int8_t field_type = self._byte()             # <<<<<<<<<<<<<<
 *         if field_type == STRUCT_END:
 *             return STRUCT_END_HEADER
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 217, __pyx_L1_error)
  __pyx_v_field_type = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":218
 *     cdef FieldHeader read_field_begin(self) except *:
 *         cdef int8_t field_type = self._byte()
 *         if field_type == STRUCT_END:             # <<<<<<<<<<<<<<
 *             return STRUCT_END_HEADER
 * 
 */
  __pyx_t_2 = ((__pyx_v_field_type == __pyx_v_8thriftrw_8protocol_6binary_STRUCT_END) != 0);
  if (__pyx_t_2) {

    /* "thriftrw/protocol/binary.pyx":219
 *         cdef int8_t field_type = self._byte()
 *         if field_type == STRUCT_END:
 *             return STRUCT_END_HEADER             # <<<<<<<<<<<<<<
 * 
 *         cdef int16_t field_id = self._i16()
 */
    __pyx_t_3 = __pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader(__pyx_v_8thriftrw_8protocol_6binary_STRUCT_END_HEADER); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 219, __pyx_L1_error)
    __pyx_r = __pyx_t_3;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":218
 *     cdef FieldHeader read_field_begin(self) except *:
 *         cdef int8_t field_type = self._byte()
 *         if field_type == STRUCT_END:             # <<<<<<<<<<<<<<
 *             return STRUCT_END_HEADER
 * 
 */
  }

  /* "thriftrw/protocol/binary.pyx":221
 *             return STRUCT_END_HEADER
 * 
 *         cdef int16_t field_id = self._i16()             # <<<<<<<<<<<<<<
 * 
 *         return FieldHeader(field_type, field_id)
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i16(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 221, __pyx_L1_error)
  __pyx_v_field_id = __pyx_t_4;

  /* "thriftrw/protocol/binary.pyx":223
 *         cdef int16_t field_id = self._i16()
 * 
 *         return FieldHeader(field_type, field_id)             # <<<<<<<<<<<<<<
 * 
 *     cdef MapHeader read_map_begin(self) except *:
 */
  __pyx_t_3.type = __pyx_v_field_type;
  __pyx_t_3.id = __pyx_v_field_id;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":216
 *         self.reader.skip(length)
 * 
 *     cdef FieldHeader read_field_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t field_type = self._byte()
 *         if field_type == STRUCT_END:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_field_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":225
 *         return FieldHeader(field_type, field_id)
 * 
 *     cdef MapHeader read_map_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t key_ttype = self._byte()
 *         cdef int8_t value_ttype = self._byte()
 */

static struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_map_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_key_ttype;
  int8_t __pyx_v_value_ttype;
  int32_t __pyx_v_length;
  struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_r;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  int32_t __pyx_t_2;
  struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_t_3;
  __Pyx_RefNannySetupContext("read_map_begin", 0);

  /* "thriftrw/protocol/binary.pyx":226
 * 
 *     cdef MapHeader read_map_begin(self) except *:
 *         cdef int8_t key_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 226, __pyx_L1_error)
  __pyx_v_key_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":227
 *     cdef MapHeader read_map_begin(self) except *:
 *         cdef int8_t key_ttype = self._byte()
 *         cdef int8_t value_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L1_error)
  __pyx_v_value_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":228
 *         cdef int8_t key_ttype = self._byte()
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 * 
 *         return MapHeader(key_ttype, value_ttype, length)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 228, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "thriftrw/protocol/binary.pyx":230
 *         cdef int32_t length = self._i32()
 * 
 *         return MapHeader(key_ttype, value_ttype, length)             # <<<<<<<<<<<<<<
 * 
 *     cdef SetHeader read_set_begin(self) except *:
 */
  __pyx_t_3.ktype = __pyx_v_key_ttype;
  __pyx_t_3.vtype = __pyx_v_value_ttype;
  __pyx_t_3.size = __pyx_v_length;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":225
 *         return FieldHeader(field_type, field_id)
 * 
 *     cdef MapHeader read_map_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t key_ttype = self._byte()
 *         cdef int8_t value_ttype = self._byte()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_map_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":232
 *         return MapHeader(key_ttype, value_ttype, length)
 * 
 *     cdef SetHeader read_set_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()
 */

static struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_set_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_value_ttype;
  int32_t __pyx_v_length;
  struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_r;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  int32_t __pyx_t_2;
  struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_t_3;
  __Pyx_RefNannySetupContext("read_set_begin", 0);

  /* "thriftrw/protocol/binary.pyx":233
 * 
 *     cdef SetHeader read_set_begin(self) except *:
 *         cdef int8_t value_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_v_value_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":234
 *     cdef SetHeader read_set_begin(self) except *:
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 * 
 *         return SetHeader(value_ttype, length)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "thriftrw/protocol/binary.pyx":236
 *         cdef int32_t length = self._i32()
 * 
 *         return SetHeader(value_ttype, length)             # <<<<<<<<<<<<<<
 * 
 *     cdef ListHeader read_list_begin(self) except *:
 */
  __pyx_t_3.type = __pyx_v_value_ttype;
  __pyx_t_3.size = __pyx_v_length;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":232
 *         return MapHeader(key_ttype, value_ttype, length)
 * 
 *     cdef SetHeader read_set_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_set_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":238
 *         return SetHeader(value_ttype, length)
 * 
 *     cdef ListHeader read_list_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()
 */

static struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_list_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_value_ttype;
  int32_t __pyx_v_length;
  struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_r;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  int32_t __pyx_t_2;
  struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_t_3;
  __Pyx_RefNannySetupContext("read_list_begin", 0);

  /* "thriftrw/protocol/binary.pyx":239
 * 
 *     cdef ListHeader read_list_begin(self) except *:
 *         cdef int8_t value_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L1_error)
  __pyx_v_value_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":240
 *     cdef ListHeader read_list_begin(self) except *:
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 * 
 *         return ListHeader(value_ttype, length)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 240, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "thriftrw/protocol/binary.pyx":242
 *         cdef int32_t length = self._i32()
 * 
 *         return ListHeader(value_ttype, length)             # <<<<<<<<<<<<<<
 * 
 *     cdef MessageHeader read_message_begin(self):
 */
  __pyx_t_3.type = __pyx_v_value_ttype;
  __pyx_t_3.size = __pyx_v_length;
  __pyx_r = __pyx_t_3;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":238
 *         return SetHeader(value_ttype, length)
 * 
 *     cdef ListHeader read_list_begin(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_list_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":244
 *         return ListHeader(value_ttype, length)
 * 
 *     cdef MessageHeader read_message_begin(self):             # <<<<<<<<<<<<<<
 *         cdef int8_t typ
 *         cdef int16_t version
 */

static struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_message_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_typ;
  int16_t __pyx_v_version;
  int32_t __pyx_v_size;
  int32_t __pyx_v_seqid;
  PyObject *__pyx_v_name = 0;
  struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int8_t __pyx_t_8;
  __Pyx_RefNannySetupContext("read_message_begin", 0);

  /* "thriftrw/protocol/binary.pyx":251
 *         cdef bytes name
 * 
 *         size = self._i32()             # <<<<<<<<<<<<<<
 *         # TODO with cython, some of the Python-specific hacks around bit
 *         # twiddling can be skipped.
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 251, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":254
 *         # TODO with cython, some of the Python-specific hacks around bit
 *         # twiddling can be skipped.
 *         if size < 0:             # <<<<<<<<<<<<<<
 *             # strict version:
 *             #
 */
  __pyx_t_2 = ((__pyx_v_size < 0) != 0);
  if (__pyx_t_2) {

    /* "thriftrw/protocol/binary.pyx":258
 *             #
 *             #     versionAndType:4 name~4 seqid:4 payload
 *             version = (size & VERSION_MASK) >> 16             # <<<<<<<<<<<<<<
 *             if version != VERSION:
 *                 raise ThriftProtocolError(
 */
    __pyx_v_version = ((__pyx_v_size & __pyx_v_8thriftrw_8protocol_6binary_VERSION_MASK) >> 16);

    /* "thriftrw/protocol/binary.pyx":259
 *             #     versionAndType:4 name~4 seqid:4 payload
 *             version = (size & VERSION_MASK) >> 16
 *             if version != VERSION:             # <<<<<<<<<<<<<<
 *                 raise ThriftProtocolError(
 *                     'Unsupported version "%r"' % version
 */
    __pyx_t_2 = ((__pyx_v_version != __pyx_v_8thriftrw_8protocol_6binary_VERSION) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "thriftrw/protocol/binary.pyx":260
 *             version = (size & VERSION_MASK) >> 16
 *             if version != VERSION:
 *                 raise ThriftProtocolError(             # <<<<<<<<<<<<<<
 *                     'Unsupported version "%r"' % version
 *                 )
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ThriftProtocolError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "thriftrw/protocol/binary.pyx":261
 *             if version != VERSION:
 *                 raise ThriftProtocolError(
 *                     'Unsupported version "%r"' % version             # <<<<<<<<<<<<<<
 *                 )
 *             typ = size & TYPE_MASK
 */
      __pyx_t_5 = __Pyx_PyInt_From_int16_t(__pyx_v_version); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_Unsupported_version_r, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
          __pyx_t_6 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 260, __pyx_L1_error)

      /* "thriftrw/protocol/binary.pyx":259
 *             #     versionAndType:4 name~4 seqid:4 payload
 *             version = (size & VERSION_MASK) >> 16
 *             if version != VERSION:             # <<<<<<<<<<<<<<
 *                 raise ThriftProtocolError(
 *                     'Unsupported version "%r"' % version
 */
    }

    /* "thriftrw/protocol/binary.pyx":263
 *                     'Unsupported version "%r"' % version
 *                 )
 *             typ = size & TYPE_MASK             # <<<<<<<<<<<<<<
 *             size = self._i32()
 *             name = self.reader.take(size)
 */
    __pyx_v_typ = (__pyx_v_size & __pyx_v_8thriftrw_8protocol_6binary_TYPE_MASK);

    /* "thriftrw/protocol/binary.pyx":264
 *                 )
 *             typ = size & TYPE_MASK
 *             size = self._i32()             # <<<<<<<<<<<<<<
 *             name = self.reader.take(size)
 *         else:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 264, __pyx_L1_error)
    __pyx_v_size = __pyx_t_1;

    /* "thriftrw/protocol/binary.pyx":265
 *             typ = size & TYPE_MASK
 *             size = self._i32()
 *             name = self.reader.take(size)             # <<<<<<<<<<<<<<
 *         else:
 *             # non-strict version:
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->take(__pyx_v_self->reader, __pyx_v_size, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "thriftrw/protocol/binary.pyx":254
 *         # TODO with cython, some of the Python-specific hacks around bit
 *         # twiddling can be skipped.
 *         if size < 0:             # <<<<<<<<<<<<<<
 *             # strict version:
 *             #
 */
    goto __pyx_L3;
  }

  /* "thriftrw/protocol/binary.pyx":270
 *             #
 *             #     name:4 type:1 seqid:4 payload
 *             name = self.reader.take(size)             # <<<<<<<<<<<<<<
 *             typ = self._byte()
 * 
 */
  /*else*/ {
    __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->take(__pyx_v_self->reader, __pyx_v_size, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "thriftrw/protocol/binary.pyx":271
 *             #     name:4 type:1 seqid:4 payload
 *             name = self.reader.take(size)
 *             typ = self._byte()             # <<<<<<<<<<<<<<
 * 
 *         seqid = self._i32()
 */
    __pyx_t_8 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 271, __pyx_L1_error)
    __pyx_v_typ = __pyx_t_8;
  }
  __pyx_L3:;

  /* "thriftrw/protocol/binary.pyx":273
 *             typ = self._byte()
 * 
 *         seqid = self._i32()             # <<<<<<<<<<<<<<
 *         return MessageHeader(name, typ, seqid)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self->__pyx_base.__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 273, __pyx_L1_error)
  __pyx_v_seqid = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":274
 * 
 *         seqid = self._i32()
 *         return MessageHeader(name, typ, seqid)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_3 = __Pyx_PyInt_From_int8_t(__pyx_v_typ); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int32_t(__pyx_v_seqid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_name);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 2, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8thriftrw_8protocol_4core_MessageHeader), __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":244
 *         return ListHeader(value_ttype, length)
 * 
 *     cdef MessageHeader read_message_begin(self):             # <<<<<<<<<<<<<<
 *         cdef int8_t typ
 *         cdef int16_t version
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.read_message_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader_2__reduce_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader_4__setstate_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolReader_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":280
 *     """Serializes values using the Thrift Binary protocol."""
 * 
 *     def __cinit__(BinaryProtocolWriter self, WriteBuffer writer):             # <<<<<<<<<<<<<<
 *         """Initialize the writer.
 * 
 */

/* Python wrapper */
static int __pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *__pyx_v_writer = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_writer,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_writer)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 280, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_writer = ((struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 280, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_writer), __pyx_ptype_8thriftrw_7_buffer_WriteBuffer, 1, "writer", 0))) __PYX_ERR(0, 280, __pyx_L1_error)
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter___cinit__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self), __pyx_v_writer);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter___cinit__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *__pyx_v_writer) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "thriftrw/protocol/binary.pyx":286
 *             WriteBuffer to which requests will be written.
 *         """
 *         self.writer = writer             # <<<<<<<<<<<<<<
 * 
 *     cdef void _write(BinaryProtocolWriter self, char* data, int length):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_writer));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_writer));
  __Pyx_GOTREF(__pyx_v_self->writer);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->writer));
  __pyx_v_self->writer = __pyx_v_writer;

  /* "thriftrw/protocol/binary.pyx":280
 *     """Serializes values using the Thrift Binary protocol."""
 * 
 *     def __cinit__(BinaryProtocolWriter self, WriteBuffer writer):             # <<<<<<<<<<<<<<
 *         """Initialize the writer.
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":288
 *         self.writer = writer
 * 
 *     cdef void _write(BinaryProtocolWriter self, char* data, int length):             # <<<<<<<<<<<<<<
 *         self.writer.write(data, length)
 * 
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter__write(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, char *__pyx_v_data, int __pyx_v_length) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_write", 0);

  /* "thriftrw/protocol/binary.pyx":289
 * 
 *     cdef void _write(BinaryProtocolWriter self, char* data, int length):
 *         self.writer.write(data, length)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_bool(self, bint value) except *:  # bool:1
 */
  ((struct __pyx_vtabstruct_8thriftrw_7_buffer_WriteBuffer *)__pyx_v_self->writer->__pyx_vtab)->write(__pyx_v_self->writer, __pyx_v_data, __pyx_v_length);

  /* "thriftrw/protocol/binary.pyx":288
 *         self.writer = writer
 * 
 *     cdef void _write(BinaryProtocolWriter self, char* data, int length):             # <<<<<<<<<<<<<<
 *         self.writer.write(data, length)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":291
 *         self.writer.write(data, length)
 * 
 *     cdef void write_bool(self, bint value) except *:  # bool:1             # <<<<<<<<<<<<<<
 *         self.write_byte(value)
 * 
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_bool(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int __pyx_v_value) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_bool", 0);

  /* "thriftrw/protocol/binary.pyx":292
 * 
 *     cdef void write_bool(self, bint value) except *:  # bool:1
 *         self.write_byte(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_byte(BinaryProtocolWriter self, int8_t value) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_value); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 292, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":291
 *         self.writer.write(data, length)
 * 
 *     cdef void write_bool(self, bint value) except *:  # bool:1             # <<<<<<<<<<<<<<
 *         self.write_byte(value)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_bool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":294
 *         self.write_byte(value)
 * 
 *     cdef void write_byte(BinaryProtocolWriter self, int8_t value) except *:             # <<<<<<<<<<<<<<
 *         # byte:1
 *         cdef char c = <char>value
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_byte(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int8_t __pyx_v_value) {
  char __pyx_v_c;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_byte", 0);

  /* "thriftrw/protocol/binary.pyx":296
 *     cdef void write_byte(BinaryProtocolWriter self, int8_t value) except *:
 *         # byte:1
 *         cdef char c = <char>value             # <<<<<<<<<<<<<<
 *         self._write(&c, 1)
 * 
 */
  __pyx_v_c = ((char)__pyx_v_value);

  /* "thriftrw/protocol/binary.pyx":297
 *         # byte:1
 *         cdef char c = <char>value
 *         self._write(&c, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_double(self, double value) except *:  # double:8
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->_write(__pyx_v_self, (&__pyx_v_c), 1);

  /* "thriftrw/protocol/binary.pyx":294
 *         self.write_byte(value)
 * 
 *     cdef void write_byte(BinaryProtocolWriter self, int8_t value) except *:             # <<<<<<<<<<<<<<
 *         # byte:1
 *         cdef char c = <char>value
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":299
 *         self._write(&c, 1)
 * 
 *     cdef void write_double(self, double value) except *:  # double:8             # <<<<<<<<<<<<<<
 *         # If confused:
 *         #
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_double(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, double __pyx_v_value) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_double", 0);

  /* "thriftrw/protocol/binary.pyx":305
 *         #
 *         # Is just "interpret the in-memory representation of value as typ"
 *         self.write_i64((<int64_t*>(&value))[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_i16(BinaryProtocolWriter self, int16_t value) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_i64(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), (((int64_t *)(&__pyx_v_value))[0])); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":299
 *         self._write(&c, 1)
 * 
 *     cdef void write_double(self, double value) except *:  # double:8             # <<<<<<<<<<<<<<
 *         # If confused:
 *         #
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":307
 *         self.write_i64((<int64_t*>(&value))[0])
 * 
 *     cdef void write_i16(BinaryProtocolWriter self, int16_t value) except *:             # <<<<<<<<<<<<<<
 *         # i16:2
 *         value = htobe16(value)
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i16(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int16_t __pyx_v_value) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_i16", 0);

  /* "thriftrw/protocol/binary.pyx":309
 *     cdef void write_i16(BinaryProtocolWriter self, int16_t value) except *:
 *         # i16:2
 *         value = htobe16(value)             # <<<<<<<<<<<<<<
 *         self._write(<char*>(&value), 2)
 * 
 */
  __pyx_v_value = htobe16(__pyx_v_value);

  /* "thriftrw/protocol/binary.pyx":310
 *         # i16:2
 *         value = htobe16(value)
 *         self._write(<char*>(&value), 2)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_i32(BinaryProtocolWriter self, int32_t value) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->_write(__pyx_v_self, ((char *)(&__pyx_v_value)), 2);

  /* "thriftrw/protocol/binary.pyx":307
 *         self.write_i64((<int64_t*>(&value))[0])
 * 
 *     cdef void write_i16(BinaryProtocolWriter self, int16_t value) except *:             # <<<<<<<<<<<<<<
 *         # i16:2
 *         value = htobe16(value)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":312
 *         self._write(<char*>(&value), 2)
 * 
 *     cdef void write_i32(BinaryProtocolWriter self, int32_t value) except *:             # <<<<<<<<<<<<<<
 *         # i32:4
 *         value = htobe32(value)
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i32(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int32_t __pyx_v_value) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_i32", 0);

  /* "thriftrw/protocol/binary.pyx":314
 *     cdef void write_i32(BinaryProtocolWriter self, int32_t value) except *:
 *         # i32:4
 *         value = htobe32(value)             # <<<<<<<<<<<<<<
 *         self._write(<char*>(&value), 4)
 * 
 */
  __pyx_v_value = htobe32(__pyx_v_value);

  /* "thriftrw/protocol/binary.pyx":315
 *         # i32:4
 *         value = htobe32(value)
 *         self._write(<char*>(&value), 4)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_i64(BinaryProtocolWriter self, int64_t value) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->_write(__pyx_v_self, ((char *)(&__pyx_v_value)), 4);

  /* "thriftrw/protocol/binary.pyx":312
 *         self._write(<char*>(&value), 2)
 * 
 *     cdef void write_i32(BinaryProtocolWriter self, int32_t value) except *:             # <<<<<<<<<<<<<<
 *         # i32:4
 *         value = htobe32(value)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":317
 *         self._write(<char*>(&value), 4)
 * 
 *     cdef void write_i64(BinaryProtocolWriter self, int64_t value) except *:             # <<<<<<<<<<<<<<
 *         # i64:8
 *         value = htobe64(value)
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i64(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, int64_t __pyx_v_value) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_i64", 0);

  /* "thriftrw/protocol/binary.pyx":319
 *     cdef void write_i64(BinaryProtocolWriter self, int64_t value) except *:
 *         # i64:8
 *         value = htobe64(value)             # <<<<<<<<<<<<<<
 *         self._write(<char*>(&value), 8)
 * 
 */
  __pyx_v_value = htobe64(__pyx_v_value);

  /* "thriftrw/protocol/binary.pyx":320
 *         # i64:8
 *         value = htobe64(value)
 *         self._write(<char*>(&value), 8)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_binary(BinaryProtocolWriter self, char* value, int32_t length) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->_write(__pyx_v_self, ((char *)(&__pyx_v_value)), 8);

  /* "thriftrw/protocol/binary.pyx":317
 *         self._write(<char*>(&value), 4)
 * 
 *     cdef void write_i64(BinaryProtocolWriter self, int64_t value) except *:             # <<<<<<<<<<<<<<
 *         # i64:8
 *         value = htobe64(value)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":322
 *         self._write(<char*>(&value), 8)
 * 
 *     cdef void write_binary(BinaryProtocolWriter self, char* value, int32_t length) except *:             # <<<<<<<<<<<<<<
 *         # len:4 str:len
 *         self.write_i32(length)
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_binary(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, char *__pyx_v_value, int32_t __pyx_v_length) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_binary", 0);

  /* "thriftrw/protocol/binary.pyx":324
 *     cdef void write_binary(BinaryProtocolWriter self, char* value, int32_t length) except *:
 *         # len:4 str:len
 *         self.write_i32(length)             # <<<<<<<<<<<<<<
 *         self._write(value, length)
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_i32(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_length); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":325
 *         # len:4 str:len
 *         self.write_i32(length)
 *         self._write(value, length)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_field_begin(BinaryProtocolWriter self,
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->_write(__pyx_v_self, __pyx_v_value, __pyx_v_length);

  /* "thriftrw/protocol/binary.pyx":322
 *         self._write(<char*>(&value), 8)
 * 
 *     cdef void write_binary(BinaryProtocolWriter self, char* value, int32_t length) except *:             # <<<<<<<<<<<<<<
 *         # len:4 str:len
 *         self.write_i32(length)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_binary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":327
 *         self._write(value, length)
 * 
 *     cdef void write_field_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                                 FieldHeader header) except *:
 *         # type:1 id:2
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_field_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_v_header) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_field_begin", 0);

  /* "thriftrw/protocol/binary.pyx":330
 *                                 FieldHeader header) except *:
 *         # type:1 id:2
 *         self.write_byte(header.type)             # <<<<<<<<<<<<<<
 *         self.write_i16(header.id)
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":331
 *         # type:1 id:2
 *         self.write_byte(header.type)
 *         self.write_i16(header.id)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_struct_end(BinaryProtocolWriter self) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_i16(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.id); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 331, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":327
 *         self._write(value, length)
 * 
 *     cdef void write_field_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                                 FieldHeader header) except *:
 *         # type:1 id:2
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_field_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":333
 *         self.write_i16(header.id)
 * 
 *     cdef void write_struct_end(BinaryProtocolWriter self) except *:             # <<<<<<<<<<<<<<
 *         self.write_byte(STRUCT_END)
 * 
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_struct_end(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_struct_end", 0);

  /* "thriftrw/protocol/binary.pyx":334
 * 
 *     cdef void write_struct_end(BinaryProtocolWriter self) except *:
 *         self.write_byte(STRUCT_END)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_map_begin(BinaryProtocolWriter self,
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_8thriftrw_8protocol_6binary_STRUCT_END); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 334, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":333
 *         self.write_i16(header.id)
 * 
 *     cdef void write_struct_end(BinaryProtocolWriter self) except *:             # <<<<<<<<<<<<<<
 *         self.write_byte(STRUCT_END)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_struct_end", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":336
 *         self.write_byte(STRUCT_END)
 * 
 *     cdef void write_map_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                               MapHeader header) except *:
 *         # key_type:1 value_type:1 count:4
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_map_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_MapHeader __pyx_v_header) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_map_begin", 0);

  /* "thriftrw/protocol/binary.pyx":339
 *                               MapHeader header) except *:
 *         # key_type:1 value_type:1 count:4
 *         self.write_byte(header.ktype)             # <<<<<<<<<<<<<<
 *         self.write_byte(header.vtype)
 *         self.write_i32(header.size)
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.ktype); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 339, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":340
 *         # key_type:1 value_type:1 count:4
 *         self.write_byte(header.ktype)
 *         self.write_byte(header.vtype)             # <<<<<<<<<<<<<<
 *         self.write_i32(header.size)
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.vtype); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 340, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":341
 *         self.write_byte(header.ktype)
 *         self.write_byte(header.vtype)
 *         self.write_i32(header.size)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_set_begin(BinaryProtocolWriter self,
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_i32(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.size); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 341, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":336
 *         self.write_byte(STRUCT_END)
 * 
 *     cdef void write_map_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                               MapHeader header) except *:
 *         # key_type:1 value_type:1 count:4
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_map_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":343
 *         self.write_i32(header.size)
 * 
 *     cdef void write_set_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                               SetHeader header) except *:
 *         # value_type:1 count:4
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_set_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_SetHeader __pyx_v_header) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_set_begin", 0);

  /* "thriftrw/protocol/binary.pyx":346
 *                               SetHeader header) except *:
 *         # value_type:1 count:4
 *         self.write_byte(header.type)             # <<<<<<<<<<<<<<
 *         self.write_i32(header.size)
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 346, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":347
 *         # value_type:1 count:4
 *         self.write_byte(header.type)
 *         self.write_i32(header.size)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_list_begin(BinaryProtocolWriter self,
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_i32(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.size); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 347, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":343
 *         self.write_i32(header.size)
 * 
 *     cdef void write_set_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                               SetHeader header) except *:
 *         # value_type:1 count:4
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_set_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":349
 *         self.write_i32(header.size)
 * 
 *     cdef void write_list_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                                ListHeader header) except *:
 *         # value_type:1 count:4
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_list_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_t_8thriftrw_8protocol_4core_ListHeader __pyx_v_header) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_list_begin", 0);

  /* "thriftrw/protocol/binary.pyx":352
 *                                ListHeader header) except *:
 *         # value_type:1 count:4
 *         self.write_byte(header.type)             # <<<<<<<<<<<<<<
 *         self.write_i32(header.size)
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":353
 *         # value_type:1 count:4
 *         self.write_byte(header.type)
 *         self.write_i32(header.size)             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_message_begin(BinaryProtocolWriter self,
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_i32(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_header.size); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 353, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":349
 *         self.write_i32(header.size)
 * 
 *     cdef void write_list_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                                ListHeader header) except *:
 *         # value_type:1 count:4
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_list_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":355
 *         self.write_i32(header.size)
 * 
 *     cdef void write_message_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                                   MessageHeader message) except *:
 *         self.write_binary(message.name, len(message.name))
 */

static void __pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_message_begin(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *__pyx_v_message) {
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  __Pyx_RefNannySetupContext("write_message_begin", 0);

  /* "thriftrw/protocol/binary.pyx":357
 *     cdef void write_message_begin(BinaryProtocolWriter self,
 *                                   MessageHeader message) except *:
 *         self.write_binary(message.name, len(message.name))             # <<<<<<<<<<<<<<
 *         self.write_byte(message.type)
 *         self.write_i32(message.seqid)
 */
  if (unlikely(__pyx_v_message->name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 357, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_AsWritableString(__pyx_v_message->name); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_message->name;
  __Pyx_INCREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 357, __pyx_L1_error)
  }
  __pyx_t_3 = PyBytes_GET_SIZE(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_binary(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_t_1, __pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":358
 *                                   MessageHeader message) except *:
 *         self.write_binary(message.name, len(message.name))
 *         self.write_byte(message.type)             # <<<<<<<<<<<<<<
 *         self.write_i32(message.seqid)
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_byte(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_message->type); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":359
 *         self.write_binary(message.name, len(message.name))
 *         self.write_byte(message.type)
 *         self.write_i32(message.seqid)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.write_i32(((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_v_self), __pyx_v_message->seqid); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":355
 *         self.write_i32(header.size)
 * 
 *     cdef void write_message_begin(BinaryProtocolWriter self,             # <<<<<<<<<<<<<<
 *                                   MessageHeader message) except *:
 *         self.write_binary(message.name, len(message.name))
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.write_message_begin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_2__reduce_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_4__setstate_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocolWriter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":365
 *     """Implements the Thrift binary protocol."""
 * 
 *     cpdef ProtocolReader reader(self, ReadBuffer buff):             # <<<<<<<<<<<<<<
 *         return BinaryProtocolReader(buff)
 * 
 */

static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_1reader(PyObject *__pyx_v_self, PyObject *__pyx_v_buff); /*proto*/
static struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_reader(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_buff, int __pyx_skip_dispatch) {
  struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("reader", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reader); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_1reader)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_buff)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_buff)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_buff)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_buff));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_buff));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_buff));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader))))) __PYX_ERR(0, 365, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":366
 * 
 *     cpdef ProtocolReader reader(self, ReadBuffer buff):
 *         return BinaryProtocolReader(buff)             # <<<<<<<<<<<<<<
 * 
 *     cpdef ProtocolWriter writer(self, WriteBuffer buff):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocolReader), ((PyObject *)__pyx_v_buff)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":365
 *     """Implements the Thrift binary protocol."""
 * 
 *     cpdef ProtocolReader reader(self, ReadBuffer buff):             # <<<<<<<<<<<<<<
 *         return BinaryProtocolReader(buff)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_1reader(PyObject *__pyx_v_self, PyObject *__pyx_v_buff); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_1reader(PyObject *__pyx_v_self, PyObject *__pyx_v_buff) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reader (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_buff), __pyx_ptype_8thriftrw_7_buffer_ReadBuffer, 1, "buff", 0))) __PYX_ERR(0, 365, __pyx_L1_error)
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_reader(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)__pyx_v_self), ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_buff));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_reader(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_buff) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("reader", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_reader(__pyx_v_self, __pyx_v_buff, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":368
 *         return BinaryProtocolReader(buff)
 * 
 *     cpdef ProtocolWriter writer(self, WriteBuffer buff):             # <<<<<<<<<<<<<<
 *         return BinaryProtocolWriter(buff)
 * 
 */

static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_3writer(PyObject *__pyx_v_self, PyObject *__pyx_v_buff); /*proto*/
static struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_writer(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *__pyx_v_buff, int __pyx_skip_dispatch) {
  struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("writer", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_writer); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_3writer)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_buff)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_buff)};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, ((PyObject *)__pyx_v_buff)};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 368, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_buff));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_buff));
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_buff));
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 368, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter))))) __PYX_ERR(0, 368, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":369
 * 
 *     cpdef ProtocolWriter writer(self, WriteBuffer buff):
 *         return BinaryProtocolWriter(buff)             # <<<<<<<<<<<<<<
 * 
 *     cpdef Message deserialize_message(self, bytes s):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocolWriter), ((PyObject *)__pyx_v_buff)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":368
 *         return BinaryProtocolReader(buff)
 * 
 *     cpdef ProtocolWriter writer(self, WriteBuffer buff):             # <<<<<<<<<<<<<<
 *         return BinaryProtocolWriter(buff)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_3writer(PyObject *__pyx_v_self, PyObject *__pyx_v_buff); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_3writer(PyObject *__pyx_v_self, PyObject *__pyx_v_buff) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writer (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_buff), __pyx_ptype_8thriftrw_7_buffer_WriteBuffer, 1, "buff", 0))) __PYX_ERR(0, 368, __pyx_L1_error)
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_2writer(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)__pyx_v_self), ((struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *)__pyx_v_buff));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_2writer(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *__pyx_v_buff) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("writer", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_writer(__pyx_v_self, __pyx_v_buff, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.writer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":371
 *         return BinaryProtocolWriter(buff)
 * 
 *     cpdef Message deserialize_message(self, bytes s):             # <<<<<<<<<<<<<<
 *         cdef ReadBuffer buff = ReadBuffer(s)
 *         return _OldBinaryProtocolReader(buff).read_message()
 */

static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_5deserialize_message(PyObject *__pyx_v_self, PyObject *__pyx_v_s); /*proto*/
static struct __pyx_obj_8thriftrw_4wire_7message_Message *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_message(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_buff = 0;
  struct __pyx_obj_8thriftrw_4wire_7message_Message *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("deserialize_message", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_deserialize_message); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_5deserialize_message)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_s};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_s};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 371, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_s);
          __Pyx_GIVEREF(__pyx_v_s);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_s);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 371, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_8thriftrw_4wire_7message_Message))))) __PYX_ERR(0, 371, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_7message_Message *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":372
 * 
 *     cpdef Message deserialize_message(self, bytes s):
 *         cdef ReadBuffer buff = ReadBuffer(s)             # <<<<<<<<<<<<<<
 *         return _OldBinaryProtocolReader(buff).read_message()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_7_buffer_ReadBuffer), __pyx_v_s); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buff = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "thriftrw/protocol/binary.pyx":373
 *     cpdef Message deserialize_message(self, bytes s):
 *         cdef ReadBuffer buff = ReadBuffer(s)
 *         return _OldBinaryProtocolReader(buff).read_message()             # <<<<<<<<<<<<<<
 * 
 *     cpdef Value deserialize_value(self, int typ, bytes s):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader), ((PyObject *)__pyx_v_buff)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_t_1)->__pyx_vtab)->read_message(((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_t_1))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_7message_Message *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":371
 *         return BinaryProtocolWriter(buff)
 * 
 *     cpdef Message deserialize_message(self, bytes s):             # <<<<<<<<<<<<<<
 *         cdef ReadBuffer buff = ReadBuffer(s)
 *         return _OldBinaryProtocolReader(buff).read_message()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.deserialize_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buff);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_5deserialize_message(PyObject *__pyx_v_self, PyObject *__pyx_v_s); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_5deserialize_message(PyObject *__pyx_v_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("deserialize_message (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_s), (&PyBytes_Type), 1, "s", 1))) __PYX_ERR(0, 371, __pyx_L1_error)
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_4deserialize_message(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)__pyx_v_self), ((PyObject*)__pyx_v_s));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_4deserialize_message(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("deserialize_message", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_message(__pyx_v_self, __pyx_v_s, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.deserialize_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":375
 *         return _OldBinaryProtocolReader(buff).read_message()
 * 
 *     cpdef Value deserialize_value(self, int typ, bytes s):             # <<<<<<<<<<<<<<
 *         cdef ReadBuffer buff = ReadBuffer(s)
 *         return _OldBinaryProtocolReader(buff).read(typ)
 */

static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_7deserialize_value(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_8thriftrw_4wire_5value_Value *__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_value(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, int __pyx_v_typ, PyObject *__pyx_v_s, int __pyx_skip_dispatch) {
  struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_buff = 0;
  struct __pyx_obj_8thriftrw_4wire_5value_Value *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("deserialize_value", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_deserialize_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_7deserialize_value)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_typ); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 375, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_s};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_v_s};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (__pyx_t_5) {
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
        __Pyx_INCREF(__pyx_v_s);
        __Pyx_GIVEREF(__pyx_v_s);
        PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_s);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_8thriftrw_4wire_5value_Value))))) __PYX_ERR(0, 375, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_Value *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":376
 * 
 *     cpdef Value deserialize_value(self, int typ, bytes s):
 *         cdef ReadBuffer buff = ReadBuffer(s)             # <<<<<<<<<<<<<<
 *         return _OldBinaryProtocolReader(buff).read(typ)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_7_buffer_ReadBuffer), __pyx_v_s); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buff = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "thriftrw/protocol/binary.pyx":377
 *     cpdef Value deserialize_value(self, int typ, bytes s):
 *         cdef ReadBuffer buff = ReadBuffer(s)
 *         return _OldBinaryProtocolReader(buff).read(typ)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader), ((PyObject *)__pyx_v_buff)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_t_1)->__pyx_vtab)->read(((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_t_1), __pyx_v_typ, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_8thriftrw_4wire_5value_Value))))) __PYX_ERR(0, 377, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_Value *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":375
 *         return _OldBinaryProtocolReader(buff).read_message()
 * 
 *     cpdef Value deserialize_value(self, int typ, bytes s):             # <<<<<<<<<<<<<<
 *         cdef ReadBuffer buff = ReadBuffer(s)
 *         return _OldBinaryProtocolReader(buff).read(typ)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.deserialize_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_buff);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_7deserialize_value(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_7deserialize_value(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_typ;
  PyObject *__pyx_v_s = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("deserialize_value (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_typ,&__pyx_n_s_s,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_typ)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_s)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("deserialize_value", 1, 2, 2, 1); __PYX_ERR(0, 375, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "deserialize_value") < 0)) __PYX_ERR(0, 375, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_typ = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_typ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 375, __pyx_L3_error)
    __pyx_v_s = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("deserialize_value", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 375, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.deserialize_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_s), (&PyBytes_Type), 1, "s", 1))) __PYX_ERR(0, 375, __pyx_L1_error)
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_6deserialize_value(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)__pyx_v_self), __pyx_v_typ, __pyx_v_s);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_6deserialize_value(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, int __pyx_v_typ, PyObject *__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("deserialize_value", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_value(__pyx_v_self, __pyx_v_typ, __pyx_v_s, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.deserialize_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef bint use_setstate
 *     state = ()
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_8__reduce_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_8__reduce_cython__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self) {
  int __pyx_v_use_setstate;
  PyObject *__pyx_v_state = NULL;
  PyObject *__pyx_v__dict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":4
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":5
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":6
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":7
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":5
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":9
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":10
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":11
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_pyx_unpickle_BinaryProtocol); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":10
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":13
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BinaryProtocol__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_pyx_unpickle_BinaryProtocol); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef bint use_setstate
 *     state = ()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":14
 *     else:
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BinaryProtocol__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_10__setstate_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_14BinaryProtocol_10__setstate_cython__(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":15
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_BinaryProtocol__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_8thriftrw_8protocol_6binary___pyx_unpickle_BinaryProtocol__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":14
 *     else:
 *         return __pyx_unpickle_BinaryProtocol, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_BinaryProtocol__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary.BinaryProtocol.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":393
 *     """
 * 
 *     def __cinit__(self, ReadBuffer reader):             # <<<<<<<<<<<<<<
 *         """Initialize the reader.
 *         :param reader:
 */

/* Python wrapper */
static int __pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_reader = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_reader,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reader)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 393, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_reader = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 393, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_reader), __pyx_ptype_8thriftrw_7_buffer_ReadBuffer, 1, "reader", 0))) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader___cinit__(((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self), __pyx_v_reader);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader___cinit__(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *__pyx_v_reader) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "thriftrw/protocol/binary.pyx":400
 *             ``num`` is negative.
 *         """
 *         self.reader = reader             # <<<<<<<<<<<<<<
 * 
 *     cdef object _reader(self, int8_t typ):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_reader));
  __Pyx_GOTREF(__pyx_v_self->reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->reader));
  __pyx_v_self->reader = __pyx_v_reader;

  /* "thriftrw/protocol/binary.pyx":393
 *     """
 * 
 *     def __cinit__(self, ReadBuffer reader):             # <<<<<<<<<<<<<<
 *         """Initialize the reader.
 *         :param reader:
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":402
 *         self.reader = reader
 * 
 *     cdef object _reader(self, int8_t typ):             # <<<<<<<<<<<<<<
 *         if typ == ttype.BOOL:
 *             return self.read_bool
 */

static PyObject *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__reader(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, int8_t __pyx_v_typ) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_reader", 0);

  /* "thriftrw/protocol/binary.pyx":403
 * 
 *     cdef object _reader(self, int8_t typ):
 *         if typ == ttype.BOOL:             # <<<<<<<<<<<<<<
 *             return self.read_bool
 *         elif typ == ttype.BYTE:
 */
  switch (__pyx_v_typ) {
    case __pyx_e_8thriftrw_4wire_5ttype_BOOL:

    /* "thriftrw/protocol/binary.pyx":404
 *     cdef object _reader(self, int8_t typ):
 *         if typ == ttype.BOOL:
 *             return self.read_bool             # <<<<<<<<<<<<<<
 *         elif typ == ttype.BYTE:
 *             return self.read_byte
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_bool); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":403
 * 
 *     cdef object _reader(self, int8_t typ):
 *         if typ == ttype.BOOL:             # <<<<<<<<<<<<<<
 *             return self.read_bool
 *         elif typ == ttype.BYTE:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":405
 *         if typ == ttype.BOOL:
 *             return self.read_bool
 *         elif typ == ttype.BYTE:             # <<<<<<<<<<<<<<
 *             return self.read_byte
 *         elif typ == ttype.DOUBLE:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_BYTE:

    /* "thriftrw/protocol/binary.pyx":406
 *             return self.read_bool
 *         elif typ == ttype.BYTE:
 *             return self.read_byte             # <<<<<<<<<<<<<<
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_byte); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":405
 *         if typ == ttype.BOOL:
 *             return self.read_bool
 *         elif typ == ttype.BYTE:             # <<<<<<<<<<<<<<
 *             return self.read_byte
 *         elif typ == ttype.DOUBLE:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":407
 *         elif typ == ttype.BYTE:
 *             return self.read_byte
 *         elif typ == ttype.DOUBLE:             # <<<<<<<<<<<<<<
 *             return self.read_double
 *         elif typ == ttype.I16:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_DOUBLE:

    /* "thriftrw/protocol/binary.pyx":408
 *             return self.read_byte
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I16:
 *             return self.read_i16
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_double); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":407
 *         elif typ == ttype.BYTE:
 *             return self.read_byte
 *         elif typ == ttype.DOUBLE:             # <<<<<<<<<<<<<<
 *             return self.read_double
 *         elif typ == ttype.I16:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":409
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double
 *         elif typ == ttype.I16:             # <<<<<<<<<<<<<<
 *             return self.read_i16
 *         elif typ == ttype.I32:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I16:

    /* "thriftrw/protocol/binary.pyx":410
 *             return self.read_double
 *         elif typ == ttype.I16:
 *             return self.read_i16             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I32:
 *             return self.read_i32
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_i16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":409
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double
 *         elif typ == ttype.I16:             # <<<<<<<<<<<<<<
 *             return self.read_i16
 *         elif typ == ttype.I32:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":411
 *         elif typ == ttype.I16:
 *             return self.read_i16
 *         elif typ == ttype.I32:             # <<<<<<<<<<<<<<
 *             return self.read_i32
 *         elif typ == ttype.I64:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I32:

    /* "thriftrw/protocol/binary.pyx":412
 *             return self.read_i16
 *         elif typ == ttype.I32:
 *             return self.read_i32             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I64:
 *             return self.read_i64
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_i32); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":411
 *         elif typ == ttype.I16:
 *             return self.read_i16
 *         elif typ == ttype.I32:             # <<<<<<<<<<<<<<
 *             return self.read_i32
 *         elif typ == ttype.I64:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":413
 *         elif typ == ttype.I32:
 *             return self.read_i32
 *         elif typ == ttype.I64:             # <<<<<<<<<<<<<<
 *             return self.read_i64
 *         elif typ == ttype.BINARY:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I64:

    /* "thriftrw/protocol/binary.pyx":414
 *             return self.read_i32
 *         elif typ == ttype.I64:
 *             return self.read_i64             # <<<<<<<<<<<<<<
 *         elif typ == ttype.BINARY:
 *             return self.read_binary
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_i64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":413
 *         elif typ == ttype.I32:
 *             return self.read_i32
 *         elif typ == ttype.I64:             # <<<<<<<<<<<<<<
 *             return self.read_i64
 *         elif typ == ttype.BINARY:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":415
 *         elif typ == ttype.I64:
 *             return self.read_i64
 *         elif typ == ttype.BINARY:             # <<<<<<<<<<<<<<
 *             return self.read_binary
 *         elif typ == ttype.STRUCT:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_BINARY:

    /* "thriftrw/protocol/binary.pyx":416
 *             return self.read_i64
 *         elif typ == ttype.BINARY:
 *             return self.read_binary             # <<<<<<<<<<<<<<
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_binary); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":415
 *         elif typ == ttype.I64:
 *             return self.read_i64
 *         elif typ == ttype.BINARY:             # <<<<<<<<<<<<<<
 *             return self.read_binary
 *         elif typ == ttype.STRUCT:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":417
 *         elif typ == ttype.BINARY:
 *             return self.read_binary
 *         elif typ == ttype.STRUCT:             # <<<<<<<<<<<<<<
 *             return self.read_struct
 *         elif typ == ttype.MAP:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_STRUCT:

    /* "thriftrw/protocol/binary.pyx":418
 *             return self.read_binary
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct             # <<<<<<<<<<<<<<
 *         elif typ == ttype.MAP:
 *             return self.read_map
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_struct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":417
 *         elif typ == ttype.BINARY:
 *             return self.read_binary
 *         elif typ == ttype.STRUCT:             # <<<<<<<<<<<<<<
 *             return self.read_struct
 *         elif typ == ttype.MAP:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":419
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct
 *         elif typ == ttype.MAP:             # <<<<<<<<<<<<<<
 *             return self.read_map
 *         elif typ == ttype.SET:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_MAP:

    /* "thriftrw/protocol/binary.pyx":420
 *             return self.read_struct
 *         elif typ == ttype.MAP:
 *             return self.read_map             # <<<<<<<<<<<<<<
 *         elif typ == ttype.SET:
 *             return self.read_set
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_map); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":419
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct
 *         elif typ == ttype.MAP:             # <<<<<<<<<<<<<<
 *             return self.read_map
 *         elif typ == ttype.SET:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":421
 *         elif typ == ttype.MAP:
 *             return self.read_map
 *         elif typ == ttype.SET:             # <<<<<<<<<<<<<<
 *             return self.read_set
 *         elif typ == ttype.LIST:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_SET:

    /* "thriftrw/protocol/binary.pyx":422
 *             return self.read_map
 *         elif typ == ttype.SET:
 *             return self.read_set             # <<<<<<<<<<<<<<
 *         elif typ == ttype.LIST:
 *             return self.read_list
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_set); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":421
 *         elif typ == ttype.MAP:
 *             return self.read_map
 *         elif typ == ttype.SET:             # <<<<<<<<<<<<<<
 *             return self.read_set
 *         elif typ == ttype.LIST:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":423
 *         elif typ == ttype.SET:
 *             return self.read_set
 *         elif typ == ttype.LIST:             # <<<<<<<<<<<<<<
 *             return self.read_list
 *         else:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_LIST:

    /* "thriftrw/protocol/binary.pyx":424
 *             return self.read_set
 *         elif typ == ttype.LIST:
 *             return self.read_list             # <<<<<<<<<<<<<<
 *         else:
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_list); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":423
 *         elif typ == ttype.SET:
 *             return self.read_set
 *         elif typ == ttype.LIST:             # <<<<<<<<<<<<<<
 *             return self.read_list
 *         else:
 */
    break;
    default:

    /* "thriftrw/protocol/binary.pyx":426
 *             return self.read_list
 *         else:
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object read(self, int8_t typ):
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ThriftProtocolError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_From_int8_t(__pyx_v_typ); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_Unknown_TType_r, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_3, __pyx_t_4};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 426, __pyx_L1_error)
    break;
  }

  /* "thriftrw/protocol/binary.pyx":402
 *         self.reader = reader
 * 
 *     cdef object _reader(self, int8_t typ):             # <<<<<<<<<<<<<<
 *         if typ == ttype.BOOL:
 *             return self.read_bool
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader._reader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":428
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)
 * 
 *     cpdef object read(self, int8_t typ):             # <<<<<<<<<<<<<<
 *         if typ == ttype.BOOL:
 *             return self.read_bool()
 */

static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_3read(PyObject *__pyx_v_self, PyObject *__pyx_arg_typ); /*proto*/
static PyObject *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, int8_t __pyx_v_typ, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("read", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_read); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_3read)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int8_t(__pyx_v_typ); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 428, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 428, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":429
 * 
 *     cpdef object read(self, int8_t typ):
 *         if typ == ttype.BOOL:             # <<<<<<<<<<<<<<
 *             return self.read_bool()
 *         elif typ == ttype.BYTE:
 */
  switch (__pyx_v_typ) {
    case __pyx_e_8thriftrw_4wire_5ttype_BOOL:

    /* "thriftrw/protocol/binary.pyx":430
 *     cpdef object read(self, int8_t typ):
 *         if typ == ttype.BOOL:
 *             return self.read_bool()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.BYTE:
 *             return self.read_byte()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_bool(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":429
 * 
 *     cpdef object read(self, int8_t typ):
 *         if typ == ttype.BOOL:             # <<<<<<<<<<<<<<
 *             return self.read_bool()
 *         elif typ == ttype.BYTE:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":431
 *         if typ == ttype.BOOL:
 *             return self.read_bool()
 *         elif typ == ttype.BYTE:             # <<<<<<<<<<<<<<
 *             return self.read_byte()
 *         elif typ == ttype.DOUBLE:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_BYTE:

    /* "thriftrw/protocol/binary.pyx":432
 *             return self.read_bool()
 *         elif typ == ttype.BYTE:
 *             return self.read_byte()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_byte(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":431
 *         if typ == ttype.BOOL:
 *             return self.read_bool()
 *         elif typ == ttype.BYTE:             # <<<<<<<<<<<<<<
 *             return self.read_byte()
 *         elif typ == ttype.DOUBLE:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":433
 *         elif typ == ttype.BYTE:
 *             return self.read_byte()
 *         elif typ == ttype.DOUBLE:             # <<<<<<<<<<<<<<
 *             return self.read_double()
 *         elif typ == ttype.I16:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_DOUBLE:

    /* "thriftrw/protocol/binary.pyx":434
 *             return self.read_byte()
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I16:
 *             return self.read_i16()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_double(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":433
 *         elif typ == ttype.BYTE:
 *             return self.read_byte()
 *         elif typ == ttype.DOUBLE:             # <<<<<<<<<<<<<<
 *             return self.read_double()
 *         elif typ == ttype.I16:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":435
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double()
 *         elif typ == ttype.I16:             # <<<<<<<<<<<<<<
 *             return self.read_i16()
 *         elif typ == ttype.I32:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I16:

    /* "thriftrw/protocol/binary.pyx":436
 *             return self.read_double()
 *         elif typ == ttype.I16:
 *             return self.read_i16()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I32:
 *             return self.read_i32()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_i16(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":435
 *         elif typ == ttype.DOUBLE:
 *             return self.read_double()
 *         elif typ == ttype.I16:             # <<<<<<<<<<<<<<
 *             return self.read_i16()
 *         elif typ == ttype.I32:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":437
 *         elif typ == ttype.I16:
 *             return self.read_i16()
 *         elif typ == ttype.I32:             # <<<<<<<<<<<<<<
 *             return self.read_i32()
 *         elif typ == ttype.I64:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I32:

    /* "thriftrw/protocol/binary.pyx":438
 *             return self.read_i16()
 *         elif typ == ttype.I32:
 *             return self.read_i32()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.I64:
 *             return self.read_i64()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_i32(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":437
 *         elif typ == ttype.I16:
 *             return self.read_i16()
 *         elif typ == ttype.I32:             # <<<<<<<<<<<<<<
 *             return self.read_i32()
 *         elif typ == ttype.I64:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":439
 *         elif typ == ttype.I32:
 *             return self.read_i32()
 *         elif typ == ttype.I64:             # <<<<<<<<<<<<<<
 *             return self.read_i64()
 *         elif typ == ttype.BINARY:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_I64:

    /* "thriftrw/protocol/binary.pyx":440
 *             return self.read_i32()
 *         elif typ == ttype.I64:
 *             return self.read_i64()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.BINARY:
 *             return self.read_binary()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_i64(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":439
 *         elif typ == ttype.I32:
 *             return self.read_i32()
 *         elif typ == ttype.I64:             # <<<<<<<<<<<<<<
 *             return self.read_i64()
 *         elif typ == ttype.BINARY:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":441
 *         elif typ == ttype.I64:
 *             return self.read_i64()
 *         elif typ == ttype.BINARY:             # <<<<<<<<<<<<<<
 *             return self.read_binary()
 *         elif typ == ttype.STRUCT:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_BINARY:

    /* "thriftrw/protocol/binary.pyx":442
 *             return self.read_i64()
 *         elif typ == ttype.BINARY:
 *             return self.read_binary()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_binary(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":441
 *         elif typ == ttype.I64:
 *             return self.read_i64()
 *         elif typ == ttype.BINARY:             # <<<<<<<<<<<<<<
 *             return self.read_binary()
 *         elif typ == ttype.STRUCT:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":443
 *         elif typ == ttype.BINARY:
 *             return self.read_binary()
 *         elif typ == ttype.STRUCT:             # <<<<<<<<<<<<<<
 *             return self.read_struct()
 *         elif typ == ttype.MAP:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_STRUCT:

    /* "thriftrw/protocol/binary.pyx":444
 *             return self.read_binary()
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.MAP:
 *             return self.read_map()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_struct(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":443
 *         elif typ == ttype.BINARY:
 *             return self.read_binary()
 *         elif typ == ttype.STRUCT:             # <<<<<<<<<<<<<<
 *             return self.read_struct()
 *         elif typ == ttype.MAP:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":445
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct()
 *         elif typ == ttype.MAP:             # <<<<<<<<<<<<<<
 *             return self.read_map()
 *         elif typ == ttype.SET:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_MAP:

    /* "thriftrw/protocol/binary.pyx":446
 *             return self.read_struct()
 *         elif typ == ttype.MAP:
 *             return self.read_map()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.SET:
 *             return self.read_set()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_map(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":445
 *         elif typ == ttype.STRUCT:
 *             return self.read_struct()
 *         elif typ == ttype.MAP:             # <<<<<<<<<<<<<<
 *             return self.read_map()
 *         elif typ == ttype.SET:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":447
 *         elif typ == ttype.MAP:
 *             return self.read_map()
 *         elif typ == ttype.SET:             # <<<<<<<<<<<<<<
 *             return self.read_set()
 *         elif typ == ttype.LIST:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_SET:

    /* "thriftrw/protocol/binary.pyx":448
 *             return self.read_map()
 *         elif typ == ttype.SET:
 *             return self.read_set()             # <<<<<<<<<<<<<<
 *         elif typ == ttype.LIST:
 *             return self.read_list()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_set(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":447
 *         elif typ == ttype.MAP:
 *             return self.read_map()
 *         elif typ == ttype.SET:             # <<<<<<<<<<<<<<
 *             return self.read_set()
 *         elif typ == ttype.LIST:
 */
    break;

    /* "thriftrw/protocol/binary.pyx":449
 *         elif typ == ttype.SET:
 *             return self.read_set()
 *         elif typ == ttype.LIST:             # <<<<<<<<<<<<<<
 *             return self.read_list()
 *         else:
 */
    case __pyx_e_8thriftrw_4wire_5ttype_LIST:

    /* "thriftrw/protocol/binary.pyx":450
 *             return self.read_set()
 *         elif typ == ttype.LIST:
 *             return self.read_list()             # <<<<<<<<<<<<<<
 *         else:
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read_list(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "thriftrw/protocol/binary.pyx":449
 *         elif typ == ttype.SET:
 *             return self.read_set()
 *         elif typ == ttype.LIST:             # <<<<<<<<<<<<<<
 *             return self.read_list()
 *         else:
 */
    break;
    default:

    /* "thriftrw/protocol/binary.pyx":452
 *             return self.read_list()
 *         else:
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _read(self, char* data, int count) except *:
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ThriftProtocolError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyInt_From_int8_t(__pyx_v_typ); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_Unknown_TType_r, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 452, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 452, __pyx_L1_error)
    break;
  }

  /* "thriftrw/protocol/binary.pyx":428
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)
 * 
 *     cpdef object read(self, int8_t typ):             # <<<<<<<<<<<<<<
 *         if typ == ttype.BOOL:
 *             return self.read_bool()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_3read(PyObject *__pyx_v_self, PyObject *__pyx_arg_typ); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_3read(PyObject *__pyx_v_self, PyObject *__pyx_arg_typ) {
  int8_t __pyx_v_typ;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read (wrapper)", 0);
  assert(__pyx_arg_typ); {
    __pyx_v_typ = __Pyx_PyInt_As_int8_t(__pyx_arg_typ); if (unlikely((__pyx_v_typ == ((int8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 428, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_2read(((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self), ((int8_t)__pyx_v_typ));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_2read(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, int8_t __pyx_v_typ) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("read", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read(__pyx_v_self, __pyx_v_typ, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":454
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)
 * 
 *     cdef void _read(self, char* data, int count) except *:             # <<<<<<<<<<<<<<
 *         self.reader.read(data, count)
 * 
 */

static void __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__read(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, char *__pyx_v_data, int __pyx_v_count) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_read", 0);

  /* "thriftrw/protocol/binary.pyx":455
 * 
 *     cdef void _read(self, char* data, int count) except *:
 *         self.reader.read(data, count)             # <<<<<<<<<<<<<<
 * 
 *     cdef int8_t _byte(self) except *:
 */
  ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->read(__pyx_v_self->reader, __pyx_v_data, __pyx_v_count, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 455, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":454
 *             raise ThriftProtocolError('Unknown TType "%r"' % typ)
 * 
 *     cdef void _read(self, char* data, int count) except *:             # <<<<<<<<<<<<<<
 *         self.reader.read(data, count)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader._read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "thriftrw/protocol/binary.pyx":457
 *         self.reader.read(data, count)
 * 
 *     cdef int8_t _byte(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char c
 *         self._read(&c, 1)
 */

static int8_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__byte(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_c;
  int8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_byte", 0);

  /* "thriftrw/protocol/binary.pyx":459
 *     cdef int8_t _byte(self) except *:
 *         cdef char c
 *         self._read(&c, 1)             # <<<<<<<<<<<<<<
 *         return <int8_t>c
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_read(__pyx_v_self, (&__pyx_v_c), 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 459, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":460
 *         cdef char c
 *         self._read(&c, 1)
 *         return <int8_t>c             # <<<<<<<<<<<<<<
 * 
 *     cdef int16_t _i16(self) except *:
 */
  __pyx_r = ((int8_t)__pyx_v_c);
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":457
 *         self.reader.read(data, count)
 * 
 *     cdef int8_t _byte(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char c
 *         self._read(&c, 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader._byte", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":462
 *         return <int8_t>c
 * 
 *     cdef int16_t _i16(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[2]
 *         self._read(data, 2)
 */

static int16_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i16(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_data[2];
  int16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_i16", 0);

  /* "thriftrw/protocol/binary.pyx":464
 *     cdef int16_t _i16(self) except *:
 *         cdef char data[2]
 *         self._read(data, 2)             # <<<<<<<<<<<<<<
 *         return be16toh((<int16_t*>data)[0])
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_read(__pyx_v_self, __pyx_v_data, 2); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":465
 *         cdef char data[2]
 *         self._read(data, 2)
 *         return be16toh((<int16_t*>data)[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef int32_t _i32(self) except *:
 */
  __pyx_r = be16toh((((int16_t *)__pyx_v_data)[0]));
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":462
 *         return <int8_t>c
 * 
 *     cdef int16_t _i16(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[2]
 *         self._read(data, 2)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader._i16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":467
 *         return be16toh((<int16_t*>data)[0])
 * 
 *     cdef int32_t _i32(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[4]
 *         self._read(data, 4)
 */

static int32_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i32(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_data[4];
  int32_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_i32", 0);

  /* "thriftrw/protocol/binary.pyx":469
 *     cdef int32_t _i32(self) except *:
 *         cdef char data[4]
 *         self._read(data, 4)             # <<<<<<<<<<<<<<
 *         return be32toh((<int32_t*>data)[0])
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_read(__pyx_v_self, __pyx_v_data, 4); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":470
 *         cdef char data[4]
 *         self._read(data, 4)
 *         return be32toh((<int32_t*>data)[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef int64_t _i64(self) except *:
 */
  __pyx_r = be32toh((((int32_t *)__pyx_v_data)[0]));
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":467
 *         return be16toh((<int16_t*>data)[0])
 * 
 *     cdef int32_t _i32(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[4]
 *         self._read(data, 4)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader._i32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":472
 *         return be32toh((<int32_t*>data)[0])
 * 
 *     cdef int64_t _i64(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[8]
 *         self._read(data, 8)
 */

static int64_t __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i64(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  char __pyx_v_data[8];
  int64_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_i64", 0);

  /* "thriftrw/protocol/binary.pyx":474
 *     cdef int64_t _i64(self) except *:
 *         cdef char data[8]
 *         self._read(data, 8)             # <<<<<<<<<<<<<<
 *         return be64toh((<int64_t*>data)[0])
 * 
 */
  ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_read(__pyx_v_self, __pyx_v_data, 8); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":475
 *         cdef char data[8]
 *         self._read(data, 8)
 *         return be64toh((<int64_t*>data)[0])             # <<<<<<<<<<<<<<
 * 
 *     cdef double _double(self) except *:
 */
  __pyx_r = be64toh((((int64_t *)__pyx_v_data)[0]));
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":472
 *         return be32toh((<int32_t*>data)[0])
 * 
 *     cdef int64_t _i64(self) except *:             # <<<<<<<<<<<<<<
 *         cdef char data[8]
 *         self._read(data, 8)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader._i64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":477
 *         return be64toh((<int64_t*>data)[0])
 * 
 *     cdef double _double(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int64_t value = self._i64()
 *         return (<double*>(&value))[0]
 */

static double __pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__double(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  int64_t __pyx_v_value;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  __Pyx_RefNannySetupContext("_double", 0);

  /* "thriftrw/protocol/binary.pyx":478
 * 
 *     cdef double _double(self) except *:
 *         cdef int64_t value = self._i64()             # <<<<<<<<<<<<<<
 *         return (<double*>(&value))[0]
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i64(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L1_error)
  __pyx_v_value = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":479
 *     cdef double _double(self) except *:
 *         cdef int64_t value = self._i64()
 *         return (<double*>(&value))[0]             # <<<<<<<<<<<<<<
 * 
 *     cdef Message read_message(self):
 */
  __pyx_r = (((double *)(&__pyx_v_value))[0]);
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":477
 *         return be64toh((<int64_t*>data)[0])
 * 
 *     cdef double _double(self) except *:             # <<<<<<<<<<<<<<
 *         cdef int64_t value = self._i64()
 *         return (<double*>(&value))[0]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader._double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":481
 *         return (<double*>(&value))[0]
 * 
 *     cdef Message read_message(self):             # <<<<<<<<<<<<<<
 *         cdef int8_t typ
 *         cdef int16_t version
 */

static struct __pyx_obj_8thriftrw_4wire_7message_Message *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_message(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_typ;
  int16_t __pyx_v_version;
  int32_t __pyx_v_size;
  PyObject *__pyx_v_name = 0;
  int32_t __pyx_v_seqid;
  PyObject *__pyx_v_body = NULL;
  struct __pyx_obj_8thriftrw_4wire_7message_Message *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int8_t __pyx_t_8;
  __Pyx_RefNannySetupContext("read_message", 0);

  /* "thriftrw/protocol/binary.pyx":487
 *         cdef bytes name
 * 
 *         size = self._i32()             # <<<<<<<<<<<<<<
 *         # TODO with cython, some of the Python-specific hacks around bit
 *         # twiddling can be skipped.
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 487, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":490
 *         # TODO with cython, some of the Python-specific hacks around bit
 *         # twiddling can be skipped.
 *         if size < 0:             # <<<<<<<<<<<<<<
 *             # strict version:
 *             #
 */
  __pyx_t_2 = ((__pyx_v_size < 0) != 0);
  if (__pyx_t_2) {

    /* "thriftrw/protocol/binary.pyx":494
 *             #
 *             #     versionAndType:4 name~4 seqid:4 payload
 *             version = (size & VERSION_MASK) >> 16             # <<<<<<<<<<<<<<
 *             if version != VERSION:
 *                 raise ThriftProtocolError(
 */
    __pyx_v_version = ((__pyx_v_size & __pyx_v_8thriftrw_8protocol_6binary_VERSION_MASK) >> 16);

    /* "thriftrw/protocol/binary.pyx":495
 *             #     versionAndType:4 name~4 seqid:4 payload
 *             version = (size & VERSION_MASK) >> 16
 *             if version != VERSION:             # <<<<<<<<<<<<<<
 *                 raise ThriftProtocolError(
 *                     'Unsupported version "%r"' % version
 */
    __pyx_t_2 = ((__pyx_v_version != __pyx_v_8thriftrw_8protocol_6binary_VERSION) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "thriftrw/protocol/binary.pyx":496
 *             version = (size & VERSION_MASK) >> 16
 *             if version != VERSION:
 *                 raise ThriftProtocolError(             # <<<<<<<<<<<<<<
 *                     'Unsupported version "%r"' % version
 *                 )
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ThriftProtocolError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "thriftrw/protocol/binary.pyx":497
 *             if version != VERSION:
 *                 raise ThriftProtocolError(
 *                     'Unsupported version "%r"' % version             # <<<<<<<<<<<<<<
 *                 )
 *             typ = size & TYPE_MASK
 */
      __pyx_t_5 = __Pyx_PyInt_From_int16_t(__pyx_v_version); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 497, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_Unsupported_version_r, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 497, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_6};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else
        #endif
        {
          __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 496, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
          __pyx_t_6 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 496, __pyx_L1_error)

      /* "thriftrw/protocol/binary.pyx":495
 *             #     versionAndType:4 name~4 seqid:4 payload
 *             version = (size & VERSION_MASK) >> 16
 *             if version != VERSION:             # <<<<<<<<<<<<<<
 *                 raise ThriftProtocolError(
 *                     'Unsupported version "%r"' % version
 */
    }

    /* "thriftrw/protocol/binary.pyx":499
 *                     'Unsupported version "%r"' % version
 *                 )
 *             typ = size & TYPE_MASK             # <<<<<<<<<<<<<<
 *             size = self._i32()
 *             name = self.reader.take(size)
 */
    __pyx_v_typ = (__pyx_v_size & __pyx_v_8thriftrw_8protocol_6binary_TYPE_MASK);

    /* "thriftrw/protocol/binary.pyx":500
 *                 )
 *             typ = size & TYPE_MASK
 *             size = self._i32()             # <<<<<<<<<<<<<<
 *             name = self.reader.take(size)
 *         else:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 500, __pyx_L1_error)
    __pyx_v_size = __pyx_t_1;

    /* "thriftrw/protocol/binary.pyx":501
 *             typ = size & TYPE_MASK
 *             size = self._i32()
 *             name = self.reader.take(size)             # <<<<<<<<<<<<<<
 *         else:
 *             # non-strict version:
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->take(__pyx_v_self->reader, __pyx_v_size, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "thriftrw/protocol/binary.pyx":490
 *         # TODO with cython, some of the Python-specific hacks around bit
 *         # twiddling can be skipped.
 *         if size < 0:             # <<<<<<<<<<<<<<
 *             # strict version:
 *             #
 */
    goto __pyx_L3;
  }

  /* "thriftrw/protocol/binary.pyx":506
 *             #
 *             #     name:4 type:1 seqid:4 payload
 *             name = self.reader.take(size)             # <<<<<<<<<<<<<<
 *             typ = self._byte()
 * 
 */
  /*else*/ {
    __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->take(__pyx_v_self->reader, __pyx_v_size, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "thriftrw/protocol/binary.pyx":507
 *             #     name:4 type:1 seqid:4 payload
 *             name = self.reader.take(size)
 *             typ = self._byte()             # <<<<<<<<<<<<<<
 * 
 *         seqid = self._i32()
 */
    __pyx_t_8 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 507, __pyx_L1_error)
    __pyx_v_typ = __pyx_t_8;
  }
  __pyx_L3:;

  /* "thriftrw/protocol/binary.pyx":509
 *             typ = self._byte()
 * 
 *         seqid = self._i32()             # <<<<<<<<<<<<<<
 *         body = self.read(ttype.STRUCT)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 509, __pyx_L1_error)
  __pyx_v_seqid = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":510
 * 
 *         seqid = self._i32()
 *         body = self.read(ttype.STRUCT)             # <<<<<<<<<<<<<<
 * 
 *         return Message(name=name, seqid=seqid, body=body, message_type=typ)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read(__pyx_v_self, __pyx_e_8thriftrw_4wire_5ttype_STRUCT, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_body = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":512
 *         body = self.read(ttype.STRUCT)
 * 
 *         return Message(name=name, seqid=seqid, body=body, message_type=typ)             # <<<<<<<<<<<<<<
 * 
 *     cdef BoolValue read_bool(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_name, __pyx_v_name) < 0) __PYX_ERR(0, 512, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_From_int32_t(__pyx_v_seqid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_seqid, __pyx_t_4) < 0) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_body, __pyx_v_body) < 0) __PYX_ERR(0, 512, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_From_int8_t(__pyx_v_typ); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_message_type, __pyx_t_4) < 0) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8thriftrw_4wire_7message_Message), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_7message_Message *)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":481
 *         return (<double*>(&value))[0]
 * 
 *     cdef Message read_message(self):             # <<<<<<<<<<<<<<
 *         cdef int8_t typ
 *         cdef int16_t version
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_message", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_body);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":514
 *         return Message(name=name, seqid=seqid, body=body, message_type=typ)
 * 
 *     cdef BoolValue read_bool(self):             # <<<<<<<<<<<<<<
 *         """Reads a boolean."""
 *         return BoolValue(self._byte() == 1)
 */

static struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_bool(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("read_bool", 0);

  /* "thriftrw/protocol/binary.pyx":516
 *     cdef BoolValue read_bool(self):
 *         """Reads a boolean."""
 *         return BoolValue(self._byte() == 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef ByteValue read_byte(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 516, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong((__pyx_t_1 == 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_BoolValue), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":514
 *         return Message(name=name, seqid=seqid, body=body, message_type=typ)
 * 
 *     cdef BoolValue read_bool(self):             # <<<<<<<<<<<<<<
 *         """Reads a boolean."""
 *         return BoolValue(self._byte() == 1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_bool", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":518
 *         return BoolValue(self._byte() == 1)
 * 
 *     cdef ByteValue read_byte(self):             # <<<<<<<<<<<<<<
 *         """Reads a byte."""
 *         return ByteValue(self._byte())
 */

static struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_byte(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("read_byte", 0);

  /* "thriftrw/protocol/binary.pyx":520
 *     cdef ByteValue read_byte(self):
 *         """Reads a byte."""
 *         return ByteValue(self._byte())             # <<<<<<<<<<<<<<
 * 
 *     cdef DoubleValue read_double(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 520, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int8_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_ByteValue), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":518
 *         return BoolValue(self._byte() == 1)
 * 
 *     cdef ByteValue read_byte(self):             # <<<<<<<<<<<<<<
 *         """Reads a byte."""
 *         return ByteValue(self._byte())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_byte", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":522
 *         return ByteValue(self._byte())
 * 
 *     cdef DoubleValue read_double(self):             # <<<<<<<<<<<<<<
 *         """Reads a double."""
 *         return DoubleValue(self._double())
 */

static struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_double(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("read_double", 0);

  /* "thriftrw/protocol/binary.pyx":524
 *     cdef DoubleValue read_double(self):
 *         """Reads a double."""
 *         return DoubleValue(self._double())             # <<<<<<<<<<<<<<
 * 
 *     cdef I16Value read_i16(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_double(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 524, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_DoubleValue), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":522
 *         return ByteValue(self._byte())
 * 
 *     cdef DoubleValue read_double(self):             # <<<<<<<<<<<<<<
 *         """Reads a double."""
 *         return DoubleValue(self._double())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_double", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":526
 *         return DoubleValue(self._double())
 * 
 *     cdef I16Value read_i16(self):             # <<<<<<<<<<<<<<
 *         """Reads a 16-bit integer."""
 *         return I16Value(self._i16())
 */

static struct __pyx_obj_8thriftrw_4wire_5value_I16Value *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i16(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj_8thriftrw_4wire_5value_I16Value *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int16_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("read_i16", 0);

  /* "thriftrw/protocol/binary.pyx":528
 *     cdef I16Value read_i16(self):
 *         """Reads a 16-bit integer."""
 *         return I16Value(self._i16())             # <<<<<<<<<<<<<<
 * 
 *     cdef I32Value read_i32(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i16(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 528, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int16_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_I16Value), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_I16Value *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":526
 *         return DoubleValue(self._double())
 * 
 *     cdef I16Value read_i16(self):             # <<<<<<<<<<<<<<
 *         """Reads a 16-bit integer."""
 *         return I16Value(self._i16())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_i16", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":530
 *         return I16Value(self._i16())
 * 
 *     cdef I32Value read_i32(self):             # <<<<<<<<<<<<<<
 *         """Reads a 32-bit integer."""
 *         return I32Value(self._i32())
 */

static struct __pyx_obj_8thriftrw_4wire_5value_I32Value *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i32(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj_8thriftrw_4wire_5value_I32Value *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("read_i32", 0);

  /* "thriftrw/protocol/binary.pyx":532
 *     cdef I32Value read_i32(self):
 *         """Reads a 32-bit integer."""
 *         return I32Value(self._i32())             # <<<<<<<<<<<<<<
 * 
 *     cdef I64Value read_i64(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 532, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int32_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_I32Value), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_I32Value *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":530
 *         return I16Value(self._i16())
 * 
 *     cdef I32Value read_i32(self):             # <<<<<<<<<<<<<<
 *         """Reads a 32-bit integer."""
 *         return I32Value(self._i32())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_i32", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":534
 *         return I32Value(self._i32())
 * 
 *     cdef I64Value read_i64(self):             # <<<<<<<<<<<<<<
 *         """Reads a 64-bit integer."""
 *         return I64Value(self._i64())
 */

static struct __pyx_obj_8thriftrw_4wire_5value_I64Value *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i64(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj_8thriftrw_4wire_5value_I64Value *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("read_i64", 0);

  /* "thriftrw/protocol/binary.pyx":536
 *     cdef I64Value read_i64(self):
 *         """Reads a 64-bit integer."""
 *         return I64Value(self._i64())             # <<<<<<<<<<<<<<
 * 
 *     cdef BinaryValue read_binary(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i64(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 536, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_I64Value), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_I64Value *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":534
 *         return I32Value(self._i32())
 * 
 *     cdef I64Value read_i64(self):             # <<<<<<<<<<<<<<
 *         """Reads a 64-bit integer."""
 *         return I64Value(self._i64())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_i64", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":538
 *         return I64Value(self._i64())
 * 
 *     cdef BinaryValue read_binary(self):             # <<<<<<<<<<<<<<
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()
 */

static struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_binary(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  int32_t __pyx_v_length;
  struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int32_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("read_binary", 0);

  /* "thriftrw/protocol/binary.pyx":540
 *     cdef BinaryValue read_binary(self):
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 *         return BinaryValue(self.reader.take(length))
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":541
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()
 *         return BinaryValue(self.reader.take(length))             # <<<<<<<<<<<<<<
 * 
 *     cdef StructValue read_struct(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer *)__pyx_v_self->reader->__pyx_vtab)->take(__pyx_v_self->reader, __pyx_v_length, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_BinaryValue), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":538
 *         return I64Value(self._i64())
 * 
 *     cdef BinaryValue read_binary(self):             # <<<<<<<<<<<<<<
 *         """Reads a binary blob."""
 *         cdef int32_t length = self._i32()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_binary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":543
 *         return BinaryValue(self.reader.take(length))
 * 
 *     cdef StructValue read_struct(self):             # <<<<<<<<<<<<<<
 *         """Reads an arbitrary Thrift struct."""
 *         fields = []
 */

static struct __pyx_obj_8thriftrw_4wire_5value_StructValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_struct(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  PyObject *__pyx_v_fields = NULL;
  int8_t __pyx_v_field_type;
  int16_t __pyx_v_field_id;
  PyObject *__pyx_v_field_value = NULL;
  struct __pyx_obj_8thriftrw_4wire_5value_StructValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int8_t __pyx_t_2;
  int __pyx_t_3;
  int16_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("read_struct", 0);

  /* "thriftrw/protocol/binary.pyx":545
 *     cdef StructValue read_struct(self):
 *         """Reads an arbitrary Thrift struct."""
 *         fields = []             # <<<<<<<<<<<<<<
 * 
 *         cdef int8_t field_type = self._byte()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fields = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "thriftrw/protocol/binary.pyx":547
 *         fields = []
 * 
 *         cdef int8_t field_type = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int16_t field_id
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 547, __pyx_L1_error)
  __pyx_v_field_type = __pyx_t_2;

  /* "thriftrw/protocol/binary.pyx":550
 *         cdef int16_t field_id
 * 
 *         while field_type != STRUCT_END:             # <<<<<<<<<<<<<<
 *             field_id = self._i16()
 *             field_value = self.read(field_type)
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_field_type != __pyx_v_8thriftrw_8protocol_6binary_STRUCT_END) != 0);
    if (!__pyx_t_3) break;

    /* "thriftrw/protocol/binary.pyx":551
 * 
 *         while field_type != STRUCT_END:
 *             field_id = self._i16()             # <<<<<<<<<<<<<<
 *             field_value = self.read(field_type)
 *             fields.append(
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i16(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 551, __pyx_L1_error)
    __pyx_v_field_id = __pyx_t_4;

    /* "thriftrw/protocol/binary.pyx":552
 *         while field_type != STRUCT_END:
 *             field_id = self._i16()
 *             field_value = self.read(field_type)             # <<<<<<<<<<<<<<
 *             fields.append(
 *                 FieldValue(
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->read(__pyx_v_self, __pyx_v_field_type, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_field_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "thriftrw/protocol/binary.pyx":555
 *             fields.append(
 *                 FieldValue(
 *                     id=field_id,             # <<<<<<<<<<<<<<
 *                     ttype=field_type,
 *                     value=field_value,
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyInt_From_int16_t(__pyx_v_field_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_id, __pyx_t_5) < 0) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "thriftrw/protocol/binary.pyx":556
 *                 FieldValue(
 *                     id=field_id,
 *                     ttype=field_type,             # <<<<<<<<<<<<<<
 *                     value=field_value,
 *                 )
 */
    __pyx_t_5 = __Pyx_PyInt_From_int8_t(__pyx_v_field_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_ttype, __pyx_t_5) < 0) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "thriftrw/protocol/binary.pyx":557
 *                     id=field_id,
 *                     ttype=field_type,
 *                     value=field_value,             # <<<<<<<<<<<<<<
 *                 )
 *             )
 */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_value, __pyx_v_field_value) < 0) __PYX_ERR(0, 555, __pyx_L1_error)

    /* "thriftrw/protocol/binary.pyx":554
 *             field_value = self.read(field_type)
 *             fields.append(
 *                 FieldValue(             # <<<<<<<<<<<<<<
 *                     id=field_id,
 *                     ttype=field_type,
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_FieldValue), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "thriftrw/protocol/binary.pyx":553
 *             field_id = self._i16()
 *             field_value = self.read(field_type)
 *             fields.append(             # <<<<<<<<<<<<<<
 *                 FieldValue(
 *                     id=field_id,
 */
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_fields, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "thriftrw/protocol/binary.pyx":561
 *             )
 * 
 *             field_type = self._byte()             # <<<<<<<<<<<<<<
 *         return StructValue(fields)
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 561, __pyx_L1_error)
    __pyx_v_field_type = __pyx_t_2;
  }

  /* "thriftrw/protocol/binary.pyx":562
 * 
 *             field_type = self._byte()
 *         return StructValue(fields)             # <<<<<<<<<<<<<<
 * 
 *     cdef MapValue read_map(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_StructValue), __pyx_v_fields); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_StructValue *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":543
 *         return BinaryValue(self.reader.take(length))
 * 
 *     cdef StructValue read_struct(self):             # <<<<<<<<<<<<<<
 *         """Reads an arbitrary Thrift struct."""
 *         fields = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_struct", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_field_value);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":564
 *         return StructValue(fields)
 * 
 *     cdef MapValue read_map(self):             # <<<<<<<<<<<<<<
 *         """Reads a map."""
 *         cdef int8_t key_ttype = self._byte()
 */

static struct __pyx_obj_8thriftrw_4wire_5value_MapValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_map(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_key_ttype;
  int8_t __pyx_v_value_ttype;
  int32_t __pyx_v_length;
  PyObject *__pyx_v_kreader = NULL;
  PyObject *__pyx_v_vreader = NULL;
  PyObject *__pyx_v_pairs = NULL;
  CYTHON_UNUSED int32_t __pyx_v_i;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  struct __pyx_obj_8thriftrw_4wire_5value_MapValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  int32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int32_t __pyx_t_4;
  int32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("read_map", 0);

  /* "thriftrw/protocol/binary.pyx":566
 *     cdef MapValue read_map(self):
 *         """Reads a map."""
 *         cdef int8_t key_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 566, __pyx_L1_error)
  __pyx_v_key_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":567
 *         """Reads a map."""
 *         cdef int8_t key_ttype = self._byte()
 *         cdef int8_t value_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L1_error)
  __pyx_v_value_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":568
 *         cdef int8_t key_ttype = self._byte()
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 * 
 *         kreader = self._reader(key_ttype)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 568, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "thriftrw/protocol/binary.pyx":570
 *         cdef int32_t length = self._i32()
 * 
 *         kreader = self._reader(key_ttype)             # <<<<<<<<<<<<<<
 *         vreader = self._reader(value_ttype)
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_reader(__pyx_v_self, __pyx_v_key_ttype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_kreader = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":571
 * 
 *         kreader = self._reader(key_ttype)
 *         vreader = self._reader(value_ttype)             # <<<<<<<<<<<<<<
 * 
 *         pairs = []
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_reader(__pyx_v_self, __pyx_v_value_ttype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_vreader = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":573
 *         vreader = self._reader(value_ttype)
 * 
 *         pairs = []             # <<<<<<<<<<<<<<
 *         for i in range(length):
 *             k = kreader(self)
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_pairs = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":574
 * 
 *         pairs = []
 *         for i in range(length):             # <<<<<<<<<<<<<<
 *             k = kreader(self)
 *             v = vreader(self)
 */
  __pyx_t_2 = __pyx_v_length;
  __pyx_t_4 = __pyx_t_2;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "thriftrw/protocol/binary.pyx":575
 *         pairs = []
 *         for i in range(length):
 *             k = kreader(self)             # <<<<<<<<<<<<<<
 *             v = vreader(self)
 *             pairs.append(MapItem(k, v))
 */
    __Pyx_INCREF(__pyx_v_kreader);
    __pyx_t_6 = __pyx_v_kreader; __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 575, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)__pyx_v_self));
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "thriftrw/protocol/binary.pyx":576
 *         for i in range(length):
 *             k = kreader(self)
 *             v = vreader(self)             # <<<<<<<<<<<<<<
 *             pairs.append(MapItem(k, v))
 * 
 */
    __Pyx_INCREF(__pyx_v_vreader);
    __pyx_t_6 = __pyx_v_vreader; __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, ((PyObject *)__pyx_v_self));
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 576, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "thriftrw/protocol/binary.pyx":577
 *             k = kreader(self)
 *             v = vreader(self)
 *             pairs.append(MapItem(k, v))             # <<<<<<<<<<<<<<
 * 
 *         return MapValue(
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_k);
    __Pyx_INCREF(__pyx_v_v);
    __Pyx_GIVEREF(__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_v);
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_MapItem), __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_pairs, __pyx_t_6); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":579
 *             pairs.append(MapItem(k, v))
 * 
 *         return MapValue(             # <<<<<<<<<<<<<<
 *             key_ttype=key_ttype,
 *             value_ttype=value_ttype,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));

  /* "thriftrw/protocol/binary.pyx":580
 * 
 *         return MapValue(
 *             key_ttype=key_ttype,             # <<<<<<<<<<<<<<
 *             value_ttype=value_ttype,
 *             pairs=pairs,
 */
  __pyx_t_6 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyInt_From_int8_t(__pyx_v_key_ttype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_key_ttype, __pyx_t_3) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":581
 *         return MapValue(
 *             key_ttype=key_ttype,
 *             value_ttype=value_ttype,             # <<<<<<<<<<<<<<
 *             pairs=pairs,
 *         )
 */
  __pyx_t_3 = __Pyx_PyInt_From_int8_t(__pyx_v_value_ttype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_value_ttype, __pyx_t_3) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":582
 *             key_ttype=key_ttype,
 *             value_ttype=value_ttype,
 *             pairs=pairs,             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_pairs, __pyx_v_pairs) < 0) __PYX_ERR(0, 580, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":579
 *             pairs.append(MapItem(k, v))
 * 
 *         return MapValue(             # <<<<<<<<<<<<<<
 *             key_ttype=key_ttype,
 *             value_ttype=value_ttype,
 */
  __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_MapValue), __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_MapValue *)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":564
 *         return StructValue(fields)
 * 
 *     cdef MapValue read_map(self):             # <<<<<<<<<<<<<<
 *         """Reads a map."""
 *         cdef int8_t key_ttype = self._byte()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_kreader);
  __Pyx_XDECREF(__pyx_v_vreader);
  __Pyx_XDECREF(__pyx_v_pairs);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":585
 *         )
 * 
 *     cdef SetValue read_set(self):             # <<<<<<<<<<<<<<
 *         """Reads a set."""
 *         cdef int8_t value_ttype = self._byte()
 */

static struct __pyx_obj_8thriftrw_4wire_5value_SetValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_set(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_value_ttype;
  int32_t __pyx_v_length;
  PyObject *__pyx_v_vreader = NULL;
  PyObject *__pyx_v_values = NULL;
  CYTHON_UNUSED int32_t __pyx_v_i;
  struct __pyx_obj_8thriftrw_4wire_5value_SetValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  int32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int32_t __pyx_t_4;
  int32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("read_set", 0);

  /* "thriftrw/protocol/binary.pyx":587
 *     cdef SetValue read_set(self):
 *         """Reads a set."""
 *         cdef int8_t value_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_v_value_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":588
 *         """Reads a set."""
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 * 
 *         vreader = self._reader(value_ttype)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 588, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "thriftrw/protocol/binary.pyx":590
 *         cdef int32_t length = self._i32()
 * 
 *         vreader = self._reader(value_ttype)             # <<<<<<<<<<<<<<
 * 
 *         values = []
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_reader(__pyx_v_self, __pyx_v_value_ttype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_vreader = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":592
 *         vreader = self._reader(value_ttype)
 * 
 *         values = []             # <<<<<<<<<<<<<<
 * 
 *         for i in range(length):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_values = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":594
 *         values = []
 * 
 *         for i in range(length):             # <<<<<<<<<<<<<<
 *             values.append(vreader(self))
 * 
 */
  __pyx_t_2 = __pyx_v_length;
  __pyx_t_4 = __pyx_t_2;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "thriftrw/protocol/binary.pyx":595
 * 
 *         for i in range(length):
 *             values.append(vreader(self))             # <<<<<<<<<<<<<<
 * 
 *         return SetValue(
 */
    __Pyx_INCREF(__pyx_v_vreader);
    __pyx_t_6 = __pyx_v_vreader; __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 595, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)__pyx_v_self));
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 595, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 595, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":597
 *             values.append(vreader(self))
 * 
 *         return SetValue(             # <<<<<<<<<<<<<<
 *             value_ttype=value_ttype,
 *             values=values
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));

  /* "thriftrw/protocol/binary.pyx":598
 * 
 *         return SetValue(
 *             value_ttype=value_ttype,             # <<<<<<<<<<<<<<
 *             values=values
 *         )
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_From_int8_t(__pyx_v_value_ttype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_value_ttype, __pyx_t_6) < 0) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "thriftrw/protocol/binary.pyx":599
 *         return SetValue(
 *             value_ttype=value_ttype,
 *             values=values             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_values, __pyx_v_values) < 0) __PYX_ERR(0, 598, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":597
 *             values.append(vreader(self))
 * 
 *         return SetValue(             # <<<<<<<<<<<<<<
 *             value_ttype=value_ttype,
 *             values=values
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_SetValue), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_SetValue *)__pyx_t_6);
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":585
 *         )
 * 
 *     cdef SetValue read_set(self):             # <<<<<<<<<<<<<<
 *         """Reads a set."""
 *         cdef int8_t value_ttype = self._byte()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vreader);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "thriftrw/protocol/binary.pyx":602
 *         )
 * 
 *     cdef ListValue read_list(self):             # <<<<<<<<<<<<<<
 *         """Reads a list."""
 *         cdef int8_t value_ttype = self._byte()
 */

static struct __pyx_obj_8thriftrw_4wire_5value_ListValue *__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_list(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  int8_t __pyx_v_value_ttype;
  int32_t __pyx_v_length;
  PyObject *__pyx_v_vreader = NULL;
  PyObject *__pyx_v_values = NULL;
  CYTHON_UNUSED int32_t __pyx_v_i;
  struct __pyx_obj_8thriftrw_4wire_5value_ListValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int8_t __pyx_t_1;
  int32_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int32_t __pyx_t_4;
  int32_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("read_list", 0);

  /* "thriftrw/protocol/binary.pyx":604
 *     cdef ListValue read_list(self):
 *         """Reads a list."""
 *         cdef int8_t value_ttype = self._byte()             # <<<<<<<<<<<<<<
 *         cdef int32_t length = self._i32()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_byte(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 604, __pyx_L1_error)
  __pyx_v_value_ttype = __pyx_t_1;

  /* "thriftrw/protocol/binary.pyx":605
 *         """Reads a list."""
 *         cdef int8_t value_ttype = self._byte()
 *         cdef int32_t length = self._i32()             # <<<<<<<<<<<<<<
 * 
 *         vreader = self._reader(value_ttype)
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_i32(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 605, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "thriftrw/protocol/binary.pyx":607
 *         cdef int32_t length = self._i32()
 * 
 *         vreader = self._reader(value_ttype)             # <<<<<<<<<<<<<<
 * 
 *         values = []
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self->__pyx_vtab)->_reader(__pyx_v_self, __pyx_v_value_ttype); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_vreader = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":609
 *         vreader = self._reader(value_ttype)
 * 
 *         values = []             # <<<<<<<<<<<<<<
 * 
 *         for i in range(length):
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_values = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "thriftrw/protocol/binary.pyx":611
 *         values = []
 * 
 *         for i in range(length):             # <<<<<<<<<<<<<<
 *             values.append(vreader(self))
 * 
 */
  __pyx_t_2 = __pyx_v_length;
  __pyx_t_4 = __pyx_t_2;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "thriftrw/protocol/binary.pyx":612
 * 
 *         for i in range(length):
 *             values.append(vreader(self))             # <<<<<<<<<<<<<<
 * 
 *         return ListValue(
 */
    __Pyx_INCREF(__pyx_v_vreader);
    __pyx_t_6 = __pyx_v_vreader; __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, ((PyObject *)__pyx_v_self)};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)__pyx_v_self));
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "thriftrw/protocol/binary.pyx":614
 *             values.append(vreader(self))
 * 
 *         return ListValue(             # <<<<<<<<<<<<<<
 *             value_ttype=value_ttype,
 *             values=values
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));

  /* "thriftrw/protocol/binary.pyx":615
 * 
 *         return ListValue(
 *             value_ttype=value_ttype,             # <<<<<<<<<<<<<<
 *             values=values
 *         )
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyInt_From_int8_t(__pyx_v_value_ttype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_value_ttype, __pyx_t_6) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "thriftrw/protocol/binary.pyx":616
 *         return ListValue(
 *             value_ttype=value_ttype,
 *             values=values             # <<<<<<<<<<<<<<
 *         )
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_values, __pyx_v_values) < 0) __PYX_ERR(0, 615, __pyx_L1_error)

  /* "thriftrw/protocol/binary.pyx":614
 *             values.append(vreader(self))
 * 
 *         return ListValue(             # <<<<<<<<<<<<<<
 *             value_ttype=value_ttype,
 *             values=values
 */
  __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8thriftrw_4wire_5value_ListValue), __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = ((struct __pyx_obj_8thriftrw_4wire_5value_ListValue *)__pyx_t_6);
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "thriftrw/protocol/binary.pyx":602
 *         )
 * 
 *     cdef ListValue read_list(self):             # <<<<<<<<<<<<<<
 *         """Reads a list."""
 *         cdef int8_t value_ttype = self._byte()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.read_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vreader);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_5__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_4__reduce_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_7__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_6__setstate_cython__(((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("thriftrw.protocol.binary._OldBinaryProtocolReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_BinaryProtocol(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */

/* Python wrapper */
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_1__pyx_unpickle_BinaryProtocol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_8thriftrw_8protocol_6binary_1__pyx_unpickle_BinaryProtocol = {"__pyx_unpickle_BinaryProtocol", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_1__pyx_unpickle_BinaryProtocol, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_8thriftrw_8protocol_6binary_1__pyx_unpickle_BinaryProtocol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_BinaryProtocol (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BinaryProtocol", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BinaryProtocol", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_BinaryProtocol") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_BinaryProtocol", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("thriftrw.protocol.binary.__pyx_unpickle_BinaryProtocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8thriftrw_8protocol_6binary___pyx_unpickle_BinaryProtocol(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8thriftrw_8protocol_6binary___pyx_unpickle_BinaryProtocol(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = NULL;
  PyObject *__pyx_v___pyx_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BinaryProtocol", 0);

  /* "(tree fragment)":2
 * def __pyx_unpickle_BinaryProtocol(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":3
 * def __pyx_unpickle_BinaryProtocol(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = BinaryProtocol.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":4
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = BinaryProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 4, __pyx_L1_error)

    /* "(tree fragment)":2
 * def __pyx_unpickle_BinaryProtocol(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":5
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = BinaryProtocol.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BinaryProtocol__set_state(<BinaryProtocol> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocol), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v___pyx_type};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v___pyx_type};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_INCREF(__pyx_v___pyx_type);
      __Pyx_GIVEREF(__pyx_v___pyx_type);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v___pyx_type);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = BinaryProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BinaryProtocol__set_state(<BinaryProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_7 = (__pyx_t_1 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":7
 *     __pyx_result = BinaryProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BinaryProtocol__set_state(<BinaryProtocol> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryProtocol__set_state(BinaryProtocol __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 7, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_8thriftrw_8protocol_6binary___pyx_unpickle_BinaryProtocol__set_state(((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = BinaryProtocol.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_BinaryProtocol__set_state(<BinaryProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":8
 *     if __pyx_state is not None:
 *         __pyx_unpickle_BinaryProtocol__set_state(<BinaryProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_BinaryProtocol__set_state(BinaryProtocol __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BinaryProtocol(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("thriftrw.protocol.binary.__pyx_unpickle_BinaryProtocol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":9
 *         __pyx_unpickle_BinaryProtocol__set_state(<BinaryProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryProtocol__set_state(BinaryProtocol __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_8thriftrw_8protocol_6binary___pyx_unpickle_BinaryProtocol__set_state(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_BinaryProtocol__set_state", 0);

  /* "(tree fragment)":10
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryProtocol__set_state(BinaryProtocol __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":11
 * cdef __pyx_unpickle_BinaryProtocol__set_state(BinaryProtocol __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 11, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":10
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryProtocol__set_state(BinaryProtocol __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":9
 *         __pyx_unpickle_BinaryProtocol__set_state(<BinaryProtocol> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_BinaryProtocol__set_state(BinaryProtocol __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("thriftrw.protocol.binary.__pyx_unpickle_BinaryProtocol__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

static struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader(PyObject *__pyx_v_obj) {
  struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_v_result;
  PyObject *__pyx_v_value = NULL;
  struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int8_t __pyx_t_10;
  int16_t __pyx_t_11;
  __Pyx_RefNannySetupContext("__pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader", 0);

  /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  __pyx_t_1 = ((!(PyMapping_Check(__pyx_v_obj) != 0)) != 0);
  if (__pyx_t_1) {

    /* "FromPyStructUtility":14
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_2 = PyErr_Format(__pyx_builtin_TypeError, ((char const *)"Expected %.16s, got %.200s"), ((char *)"a mapping"), Py_TYPE(__pyx_v_obj)->tp_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":13
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader(obj) except *:
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):             # <<<<<<<<<<<<<<
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 */
  }

  /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['type']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "FromPyStructUtility":17
 * 
 *     try:
 *         value = obj['type']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'type'")
 */
      __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 17, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_value = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['type']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "FromPyStructUtility":18
 *     try:
 *         value = obj['type']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'type'")
 *     result.type = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(1, 18, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "FromPyStructUtility":19
 *         value = obj['type']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'type'")             # <<<<<<<<<<<<<<
 *     result.type = value
 *     try:
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 19, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(1, 19, __pyx_L6_except_error)
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "FromPyStructUtility":16
 *         PyErr_Format(TypeError, b"Expected %.16s, got %.200s", b"a mapping", Py_TYPE(obj).tp_name)
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['type']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L9_try_end:;
  }

  /* "FromPyStructUtility":20
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'type'")
 *     result.type = value             # <<<<<<<<<<<<<<
 *     try:
 *         value = obj['id']
 */
  __pyx_t_10 = __Pyx_PyInt_As_int8_t(__pyx_v_value); if (unlikely((__pyx_t_10 == ((int8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_v_result.type = __pyx_t_10;

  /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'type'")
 *     result.type = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['id']
 *     except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "FromPyStructUtility":22
 *     result.type = value
 *     try:
 *         value = obj['id']             # <<<<<<<<<<<<<<
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'id'")
 */
      __pyx_t_8 = __Pyx_PyObject_Dict_GetItem(__pyx_v_obj, __pyx_n_s_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 22, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'type'")
 *     result.type = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['id']
 *     except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L17_try_end;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "FromPyStructUtility":23
 *     try:
 *         value = obj['id']
 *     except KeyError:             # <<<<<<<<<<<<<<
 *         raise ValueError("No value specified for struct attribute 'id'")
 *     result.id = value
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(1, 23, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);

      /* "FromPyStructUtility":24
 *         value = obj['id']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'id'")             # <<<<<<<<<<<<<<
 *     result.id = value
 *     return result
 */
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 24, __pyx_L14_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(1, 24, __pyx_L14_except_error)
    }
    goto __pyx_L14_except_error;
    __pyx_L14_except_error:;

    /* "FromPyStructUtility":21
 *         raise ValueError("No value specified for struct attribute 'type'")
 *     result.type = value
 *     try:             # <<<<<<<<<<<<<<
 *         value = obj['id']
 *     except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L17_try_end:;
  }

  /* "FromPyStructUtility":25
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'id'")
 *     result.id = value             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_11 = __Pyx_PyInt_As_int16_t(__pyx_v_value); if (unlikely((__pyx_t_11 == ((int16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 25, __pyx_L1_error)
  __pyx_v_result.id = __pyx_t_11;

  /* "FromPyStructUtility":26
 *         raise ValueError("No value specified for struct attribute 'id'")
 *     result.id = value
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "FromPyStructUtility":11
 * 
 * @cname("__pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader")
 * cdef struct_type __pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader(obj) except *:             # <<<<<<<<<<<<<<
 *     cdef struct_type result
 *     if not PyMapping_Check(obj):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("FromPyStructUtility.__pyx_convert__from_py_struct____pyx_t_8thriftrw_8protocol_4core_FieldHeader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(BoolValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(BoolValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(BoolValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 */

static PyObject *__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(BoolValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_58__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_BoolValue______OldBi, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(BoolValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(ByteValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(ByteValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(ByteValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'"""
 */

static PyObject *__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(ByteValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_58__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_ByteValue______OldBi, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(ByteValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ByteValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(DoubleValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(DoubleValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(DoubleValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'"""
 */

static PyObject *__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(DoubleValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_60__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_DoubleValue______Old, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__14)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(DoubleValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'DoubleValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(I16Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(I16Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(I16Value (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'"""
 */

static PyObject *__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_I16Value *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(I16Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_I16Value______OldBin, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__16)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(I16Value (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I16Value'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(I32Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(I32Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(I32Value (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'"""
 */

static PyObject *__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_I32Value *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(I32Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_I32Value______OldBin, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(I32Value (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I32Value'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(I64Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(I64Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(I64Value (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'"""
 */

static PyObject *__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_I64Value *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(I64Value (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_I64Value______OldBin, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(I64Value (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'I64Value'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(BinaryValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(BinaryValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(BinaryValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'"""
 */

static PyObject *__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(BinaryValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_60__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_BinaryValue______Old, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(BinaryValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BinaryValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(StructValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(StructValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(StructValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'"""
 */

static PyObject *__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_StructValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(StructValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_60__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_StructValue______Old, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(StructValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'StructValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(MapValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(MapValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(MapValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'"""
 */

static PyObject *__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_MapValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(MapValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_MapValue______OldBin, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(MapValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'MapValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(SetValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(SetValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(SetValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'"""
 */

static PyObject *__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_SetValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(SetValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_57__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_SetValue______OldBin, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(SetValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'SetValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(ListValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 *         return f(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_wrap[] = "wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_self), __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, 1, "self", 0))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_wrap(__pyx_self, ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)__pyx_v_self));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_wrap(PyObject *__pyx_self, struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *__pyx_v_self) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 *         return f(self)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_f(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(ListValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 *         return f(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(ListValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 */

static PyObject *__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(struct __pyx_obj_8thriftrw_4wire_5value_ListValue *(*__pyx_v_f)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(1, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(ListValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 *         return f(self)
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_11cfunc_dot_to_py_58__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_ListValue______OldBi, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 *         return f(self)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(ListValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocol __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocol;

static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocol(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *p;
  PyObject *o = __pyx_ptype_8thriftrw_8protocol_4core_Protocol->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8thriftrw_8protocol_4core_Protocol*)__pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocol;
  return o;
}

static void __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocol(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  if (likely(__pyx_ptype_8thriftrw_8protocol_4core_Protocol)) __pyx_ptype_8thriftrw_8protocol_4core_Protocol->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocol);
}

static PyMethodDef __pyx_methods_8thriftrw_8protocol_6binary_BinaryProtocol[] = {
  {"reader", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_1reader, METH_O, 0},
  {"writer", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_3writer, METH_O, 0},
  {"deserialize_message", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_5deserialize_message, METH_O, 0},
  {"deserialize_value", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_7deserialize_value, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_14BinaryProtocol_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.BinaryProtocol", /*tp_name*/
  sizeof(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocol), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocol, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Implements the Thrift binary protocol.", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8thriftrw_8protocol_6binary_BinaryProtocol, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocol, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolReader __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader;

static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocolReader(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *p;
  PyObject *o = __pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolReader*)__pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolReader;
  p->reader = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocolReader(PyObject *o) {
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *p = (struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->reader);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader)) __pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocolReader);
}

static int __pyx_tp_traverse_8thriftrw_8protocol_6binary_BinaryProtocolReader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *p = (struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *)o;
  e = ((likely(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader)) ? ((__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader->tp_traverse) ? __pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_8thriftrw_8protocol_6binary_BinaryProtocolReader)); if (e) return e;
  if (p->reader) {
    e = (*v)(((PyObject *)p->reader), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8thriftrw_8protocol_6binary_BinaryProtocolReader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *p = (struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *)o;
  if (likely(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader)) { if (__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader->tp_clear) __pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_8thriftrw_8protocol_6binary_BinaryProtocolReader);
  tmp = ((PyObject*)p->reader);
  p->reader = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8thriftrw_8protocol_6binary_BinaryProtocolReader[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolReader_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.BinaryProtocolReader", /*tp_name*/
  sizeof(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocolReader, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Parser for the binary protocol.", /*tp_doc*/
  __pyx_tp_traverse_8thriftrw_8protocol_6binary_BinaryProtocolReader, /*tp_traverse*/
  __pyx_tp_clear_8thriftrw_8protocol_6binary_BinaryProtocolReader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8thriftrw_8protocol_6binary_BinaryProtocolReader, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocolReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary_BinaryProtocolWriter __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter;

static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocolWriter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *p;
  PyObject *o = __pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolWriter*)__pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolWriter;
  p->writer = ((struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocolWriter(PyObject *o) {
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *p = (struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->writer);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter)) __pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocolWriter);
}

static int __pyx_tp_traverse_8thriftrw_8protocol_6binary_BinaryProtocolWriter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *p = (struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)o;
  e = ((likely(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter)) ? ((__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter->tp_traverse) ? __pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_8thriftrw_8protocol_6binary_BinaryProtocolWriter)); if (e) return e;
  if (p->writer) {
    e = (*v)(((PyObject *)p->writer), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8thriftrw_8protocol_6binary_BinaryProtocolWriter(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *p = (struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *)o;
  if (likely(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter)) { if (__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter->tp_clear) __pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_8thriftrw_8protocol_6binary_BinaryProtocolWriter);
  tmp = ((PyObject*)p->writer);
  p->writer = ((struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8thriftrw_8protocol_6binary_BinaryProtocolWriter[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.BinaryProtocolWriter", /*tp_name*/
  sizeof(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8thriftrw_8protocol_6binary_BinaryProtocolWriter, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Serializes values using the Thrift Binary protocol.", /*tp_doc*/
  __pyx_tp_traverse_8thriftrw_8protocol_6binary_BinaryProtocolWriter, /*tp_traverse*/
  __pyx_tp_clear_8thriftrw_8protocol_6binary_BinaryProtocolWriter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8thriftrw_8protocol_6binary_BinaryProtocolWriter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8thriftrw_8protocol_6binary_BinaryProtocolWriter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8thriftrw_8protocol_6binary__OldBinaryProtocolReader __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader;

static PyObject *__pyx_tp_new_8thriftrw_8protocol_6binary__OldBinaryProtocolReader(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)o);
  p->__pyx_vtab = __pyx_vtabptr_8thriftrw_8protocol_6binary__OldBinaryProtocolReader;
  p->reader = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_8thriftrw_8protocol_6binary__OldBinaryProtocolReader(PyObject *o) {
  struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *p = (struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->reader);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8thriftrw_8protocol_6binary__OldBinaryProtocolReader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *p = (struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)o;
  if (p->reader) {
    e = (*v)(((PyObject *)p->reader), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8thriftrw_8protocol_6binary__OldBinaryProtocolReader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *p = (struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *)o;
  tmp = ((PyObject*)p->reader);
  p->reader = ((struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8thriftrw_8protocol_6binary__OldBinaryProtocolReader[] = {
  {"read", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_3read, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_5__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_7__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary._OldBinaryProtocolReader", /*tp_name*/
  sizeof(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Old version of BinaryProtocolReader that returned Value representations.\n\n    Deserialization was refactored in 1.4 to avoid this step as an optimization.\n    This class exists to provide support for backwards compatibility but will\n    be removed in a later version.\n\n    .. deprecated:: 1.4\n    ", /*tp_doc*/
  __pyx_tp_traverse_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, /*tp_traverse*/
  __pyx_tp_clear_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8thriftrw_8protocol_6binary__OldBinaryProtocolReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "thriftrw.protocol.binary.__pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_binary(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_binary},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "binary",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_u_BinaryProtocol, __pyx_k_BinaryProtocol, sizeof(__pyx_k_BinaryProtocol), 0, 1, 0, 1},
  {&__pyx_n_s_EndOfInputError, __pyx_k_EndOfInputError, sizeof(__pyx_k_EndOfInputError), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_s_No_value_specified_for_struct_at, __pyx_k_No_value_specified_for_struct_at, sizeof(__pyx_k_No_value_specified_for_struct_at), 0, 0, 1, 0},
  {&__pyx_kp_s_No_value_specified_for_struct_at_2, __pyx_k_No_value_specified_for_struct_at_2, sizeof(__pyx_k_No_value_specified_for_struct_at_2), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_BinaryValue______Old, __pyx_k_Pyx_CFunc_BinaryValue______Old, sizeof(__pyx_k_Pyx_CFunc_BinaryValue______Old), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_BoolValue______OldBi, __pyx_k_Pyx_CFunc_BoolValue______OldBi, sizeof(__pyx_k_Pyx_CFunc_BoolValue______OldBi), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_ByteValue______OldBi, __pyx_k_Pyx_CFunc_ByteValue______OldBi, sizeof(__pyx_k_Pyx_CFunc_ByteValue______OldBi), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_DoubleValue______Old, __pyx_k_Pyx_CFunc_DoubleValue______Old, sizeof(__pyx_k_Pyx_CFunc_DoubleValue______Old), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_I16Value______OldBin, __pyx_k_Pyx_CFunc_I16Value______OldBin, sizeof(__pyx_k_Pyx_CFunc_I16Value______OldBin), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_I32Value______OldBin, __pyx_k_Pyx_CFunc_I32Value______OldBin, sizeof(__pyx_k_Pyx_CFunc_I32Value______OldBin), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_I64Value______OldBin, __pyx_k_Pyx_CFunc_I64Value______OldBin, sizeof(__pyx_k_Pyx_CFunc_I64Value______OldBin), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_ListValue______OldBi, __pyx_k_Pyx_CFunc_ListValue______OldBi, sizeof(__pyx_k_Pyx_CFunc_ListValue______OldBi), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_MapValue______OldBin, __pyx_k_Pyx_CFunc_MapValue______OldBin, sizeof(__pyx_k_Pyx_CFunc_MapValue______OldBin), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_SetValue______OldBin, __pyx_k_Pyx_CFunc_SetValue______OldBin, sizeof(__pyx_k_Pyx_CFunc_SetValue______OldBin), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_StructValue______Old, __pyx_k_Pyx_CFunc_StructValue______Old, sizeof(__pyx_k_Pyx_CFunc_StructValue______Old), 0, 0, 1, 1},
  {&__pyx_n_s_ThriftProtocolError, __pyx_k_ThriftProtocolError, sizeof(__pyx_k_ThriftProtocolError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_u_Unknown_TType_r, __pyx_k_Unknown_TType_r, sizeof(__pyx_k_Unknown_TType_r), 0, 1, 0, 0},
  {&__pyx_kp_u_Unsupported_version_r, __pyx_k_Unsupported_version_r, sizeof(__pyx_k_Unsupported_version_r), 0, 1, 0, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_body, __pyx_k_body, sizeof(__pyx_k_body), 0, 0, 1, 1},
  {&__pyx_n_s_cfunc_to_py, __pyx_k_cfunc_to_py, sizeof(__pyx_k_cfunc_to_py), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_deserialize_message, __pyx_k_deserialize_message, sizeof(__pyx_k_deserialize_message), 0, 0, 1, 1},
  {&__pyx_n_s_deserialize_value, __pyx_k_deserialize_value, sizeof(__pyx_k_deserialize_value), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_errors, __pyx_k_errors, sizeof(__pyx_k_errors), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_key_ttype, __pyx_k_key_ttype, sizeof(__pyx_k_key_ttype), 0, 0, 1, 1},
  {&__pyx_n_s_ktype, __pyx_k_ktype, sizeof(__pyx_k_ktype), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_message_type, __pyx_k_message_type, sizeof(__pyx_k_message_type), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_pairs, __pyx_k_pairs, sizeof(__pyx_k_pairs), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_BinaryProtocol, __pyx_k_pyx_unpickle_BinaryProtocol, sizeof(__pyx_k_pyx_unpickle_BinaryProtocol), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_reader, __pyx_k_reader, sizeof(__pyx_k_reader), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_seqid, __pyx_k_seqid, sizeof(__pyx_k_seqid), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_thriftrw_protocol_binary, __pyx_k_thriftrw_protocol_binary, sizeof(__pyx_k_thriftrw_protocol_binary), 0, 0, 1, 1},
  {&__pyx_n_s_ttype, __pyx_k_ttype, sizeof(__pyx_k_ttype), 0, 0, 1, 1},
  {&__pyx_n_s_typ, __pyx_k_typ, sizeof(__pyx_k_typ), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_value_ttype, __pyx_k_value_ttype, sizeof(__pyx_k_value_ttype), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_vtype, __pyx_k_vtype, sizeof(__pyx_k_vtype), 0, 0, 1, 1},
  {&__pyx_n_s_wrap, __pyx_k_wrap, sizeof(__pyx_k_wrap), 0, 0, 1, 1},
  {&__pyx_n_s_writer, __pyx_k_writer, sizeof(__pyx_k_writer), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 138, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(1, 19, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "FromPyStructUtility":19
 *         value = obj['type']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'type'")             # <<<<<<<<<<<<<<
 *     result.type = value
 *     try:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "FromPyStructUtility":24
 *         value = obj['id']
 *     except KeyError:
 *         raise ValueError("No value specified for struct attribute 'id'")             # <<<<<<<<<<<<<<
 *     result.id = value
 *     return result
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_No_value_specified_for_struct_at_2); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py(BoolValue (*f)(_OldBinaryProtocolReader) ):
 *     def wrap(_OldBinaryProtocolReader self):             # <<<<<<<<<<<<<<
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'BoolValue'"""
 *         return f(self)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__9, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__11, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__13, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(1, 65, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_BinaryProtocol(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */
  __pyx_tuple__31 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_BinaryProtocol, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_8thriftrw_8protocol_6binary_STRUCT_END_HEADER = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_ptype_8thriftrw_8protocol_4core_Protocol = __Pyx_ImportType("thriftrw.protocol.core", "Protocol", sizeof(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol), 1); if (unlikely(!__pyx_ptype_8thriftrw_8protocol_4core_Protocol)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_8protocol_4core_Protocol = (struct __pyx_vtabstruct_8thriftrw_8protocol_4core_Protocol*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_8protocol_4core_Protocol->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_8protocol_4core_Protocol)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocol = &__pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocol;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocol.__pyx_base = *__pyx_vtabptr_8thriftrw_8protocol_4core_Protocol;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocol.__pyx_base.writer = (struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *(*)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer *, int __pyx_skip_dispatch))__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_writer;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocol.__pyx_base.reader = (struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *(*)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer *, int __pyx_skip_dispatch))__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_reader;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocol.__pyx_base.deserialize_value = (struct __pyx_obj_8thriftrw_4wire_5value_Value *(*)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, int, PyObject *, int __pyx_skip_dispatch))__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_value;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocol.__pyx_base.deserialize_message = (struct __pyx_obj_8thriftrw_4wire_7message_Message *(*)(struct __pyx_obj_8thriftrw_8protocol_4core_Protocol *, PyObject *, int __pyx_skip_dispatch))__pyx_f_8thriftrw_8protocol_6binary_14BinaryProtocol_deserialize_message;
  __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol.tp_base = __pyx_ptype_8thriftrw_8protocol_4core_Protocol;
  if (PyType_Ready(&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
  __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol.tp_dictoffset && __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol.tp_dict, __pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocol) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "BinaryProtocol", (PyObject *)&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol) < 0) __PYX_ERR(0, 362, __pyx_L1_error)
  __pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocol = &__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocol;
  __pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader = __Pyx_ImportType("thriftrw.protocol.core", "ProtocolReader", sizeof(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader), 1); if (unlikely(!__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolReader = (struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolReader*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolReader)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolReader = &__pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base = *__pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolReader;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.skip = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *, int))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.skip_binary = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_binary;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.skip_map = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_map;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.skip_list = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_list;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.skip_set = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_set;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.skip_struct = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_skip_struct;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_bool = (int (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_bool;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_byte = (int8_t (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_byte;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_double = (double (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_double;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_i16 = (int16_t (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i16;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_i32 = (int32_t (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i32;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_i64 = (int64_t (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_i64;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_binary = (PyObject *(*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_binary;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_field_begin = (struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_field_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_map_begin = (struct __pyx_t_8thriftrw_8protocol_4core_MapHeader (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_map_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_set_begin = (struct __pyx_t_8thriftrw_8protocol_4core_SetHeader (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_set_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_list_begin = (struct __pyx_t_8thriftrw_8protocol_4core_ListHeader (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_list_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader.__pyx_base.read_message_begin = (struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *(*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader_read_message_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader._read = (void (*)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *, char *, int))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__read;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader._byte = (int8_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__byte;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader._i16 = (int16_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i16;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader._i32 = (int32_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i32;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader._i64 = (int64_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__i64;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolReader._double = (double (*)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolReader__double;
  __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader.tp_base = __pyx_ptype_8thriftrw_8protocol_4core_ProtocolReader;
  if (PyType_Ready(&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader.tp_dictoffset && __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader.tp_dict, __pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolReader) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "BinaryProtocolReader", (PyObject *)&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader) < 0) __PYX_ERR(0, 91, __pyx_L1_error)
  __pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocolReader = &__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolReader;
  __pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter = __Pyx_ImportType("thriftrw.protocol.core", "ProtocolWriter", sizeof(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter), 1); if (unlikely(!__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolWriter = (struct __pyx_vtabstruct_8thriftrw_8protocol_4core_ProtocolWriter*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolWriter)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolWriter = &__pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base = *__pyx_vtabptr_8thriftrw_8protocol_4core_ProtocolWriter;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_bool = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_bool;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_byte = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int8_t))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_byte;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_double = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, double))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_double;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_i16 = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int16_t))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i16;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_i32 = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int32_t))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i32;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_i64 = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, int64_t))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_i64;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_binary = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, char *, int32_t))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_binary;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_field_begin = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_FieldHeader))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_field_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_struct_end = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_struct_end;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_map_begin = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_MapHeader))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_map_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_set_begin = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_SetHeader))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_set_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_list_begin = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_t_8thriftrw_8protocol_4core_ListHeader))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_list_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter.__pyx_base.write_message_begin = (void (*)(struct __pyx_obj_8thriftrw_8protocol_4core_ProtocolWriter *, struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader *))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter_write_message_begin;
  __pyx_vtable_8thriftrw_8protocol_6binary_BinaryProtocolWriter._write = (void (*)(struct __pyx_obj_8thriftrw_8protocol_6binary_BinaryProtocolWriter *, char *, int))__pyx_f_8thriftrw_8protocol_6binary_20BinaryProtocolWriter__write;
  __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter.tp_base = __pyx_ptype_8thriftrw_8protocol_4core_ProtocolWriter;
  if (PyType_Ready(&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
  __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter.tp_dictoffset && __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter.tp_dict, __pyx_vtabptr_8thriftrw_8protocol_6binary_BinaryProtocolWriter) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "BinaryProtocolWriter", (PyObject *)&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
  __pyx_ptype_8thriftrw_8protocol_6binary_BinaryProtocolWriter = &__pyx_type_8thriftrw_8protocol_6binary_BinaryProtocolWriter;
  __pyx_vtabptr_8thriftrw_8protocol_6binary__OldBinaryProtocolReader = &__pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader._reader = (PyObject *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *, int8_t))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__reader;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read = (PyObject *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *, int8_t, int __pyx_skip_dispatch))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader._read = (void (*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *, char *, int))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__read;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader._byte = (int8_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__byte;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader._i16 = (int16_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i16;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader._i32 = (int32_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i32;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader._i64 = (int64_t (*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__i64;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader._double = (double (*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader__double;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_message = (struct __pyx_obj_8thriftrw_4wire_7message_Message *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_message;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_bool = (struct __pyx_obj_8thriftrw_4wire_5value_BoolValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_bool;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_byte = (struct __pyx_obj_8thriftrw_4wire_5value_ByteValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_byte;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_double = (struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_double;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_i16 = (struct __pyx_obj_8thriftrw_4wire_5value_I16Value *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i16;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_i32 = (struct __pyx_obj_8thriftrw_4wire_5value_I32Value *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i32;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_i64 = (struct __pyx_obj_8thriftrw_4wire_5value_I64Value *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_i64;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_binary = (struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_binary;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_struct = (struct __pyx_obj_8thriftrw_4wire_5value_StructValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_struct;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_map = (struct __pyx_obj_8thriftrw_4wire_5value_MapValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_map;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_set = (struct __pyx_obj_8thriftrw_4wire_5value_SetValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_set;
  __pyx_vtable_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.read_list = (struct __pyx_obj_8thriftrw_4wire_5value_ListValue *(*)(struct __pyx_obj_8thriftrw_8protocol_6binary__OldBinaryProtocolReader *))__pyx_f_8thriftrw_8protocol_6binary_24_OldBinaryProtocolReader_read_list;
  if (PyType_Ready(&__pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader) < 0) __PYX_ERR(0, 383, __pyx_L1_error)
  __pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.tp_dictoffset && __pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader.tp_dict, __pyx_vtabptr_8thriftrw_8protocol_6binary__OldBinaryProtocolReader) < 0) __PYX_ERR(0, 383, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "_OldBinaryProtocolReader", (PyObject *)&__pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader) < 0) __PYX_ERR(0, 383, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader) < 0) __PYX_ERR(0, 383, __pyx_L1_error)
  __pyx_ptype_8thriftrw_8protocol_6binary__OldBinaryProtocolReader = &__pyx_type_8thriftrw_8protocol_6binary__OldBinaryProtocolReader;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_BoolValue______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_ByteValue______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_DoubleValue______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_I16Value______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_I32Value______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_I64Value______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_BinaryValue______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_StructValue______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_MapValue______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_SetValue______OldBinaryProtocolReader___to_py;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py = &__pyx_scope_struct____Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_ptype_8thriftrw_4wire_5value_ValueVisitor = __Pyx_ImportType("thriftrw.wire.value", "ValueVisitor", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_ValueVisitor), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_ValueVisitor)) __PYX_ERR(2, 31, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_ValueVisitor = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_ValueVisitor*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_ValueVisitor->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_ValueVisitor)) __PYX_ERR(2, 31, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value__ValueVisitorArgs = __Pyx_ImportType("thriftrw.wire.value", "_ValueVisitorArgs", sizeof(struct __pyx_obj_8thriftrw_4wire_5value__ValueVisitorArgs), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value__ValueVisitorArgs)) __PYX_ERR(2, 56, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value__ValueVisitorArgs = (struct __pyx_vtabstruct_8thriftrw_4wire_5value__ValueVisitorArgs*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value__ValueVisitorArgs->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value__ValueVisitorArgs)) __PYX_ERR(2, 56, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_Value = __Pyx_ImportType("thriftrw.wire.value", "Value", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_Value), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_Value)) __PYX_ERR(2, 60, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_Value = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_Value*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_Value->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_Value)) __PYX_ERR(2, 60, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_BoolValue = __Pyx_ImportType("thriftrw.wire.value", "BoolValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_BoolValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_BoolValue)) __PYX_ERR(2, 65, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_BoolValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_BoolValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_BoolValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_BoolValue)) __PYX_ERR(2, 65, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_ByteValue = __Pyx_ImportType("thriftrw.wire.value", "ByteValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_ByteValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_ByteValue)) __PYX_ERR(2, 69, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_ByteValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_ByteValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_ByteValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_ByteValue)) __PYX_ERR(2, 69, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_DoubleValue = __Pyx_ImportType("thriftrw.wire.value", "DoubleValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_DoubleValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_DoubleValue)) __PYX_ERR(2, 73, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_DoubleValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_DoubleValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_DoubleValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_DoubleValue)) __PYX_ERR(2, 73, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_I16Value = __Pyx_ImportType("thriftrw.wire.value", "I16Value", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_I16Value), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_I16Value)) __PYX_ERR(2, 77, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_I16Value = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_I16Value*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_I16Value->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_I16Value)) __PYX_ERR(2, 77, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_I32Value = __Pyx_ImportType("thriftrw.wire.value", "I32Value", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_I32Value), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_I32Value)) __PYX_ERR(2, 81, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_I32Value = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_I32Value*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_I32Value->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_I32Value)) __PYX_ERR(2, 81, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_I64Value = __Pyx_ImportType("thriftrw.wire.value", "I64Value", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_I64Value), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_I64Value)) __PYX_ERR(2, 85, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_I64Value = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_I64Value*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_I64Value->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_I64Value)) __PYX_ERR(2, 85, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_BinaryValue = __Pyx_ImportType("thriftrw.wire.value", "BinaryValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_BinaryValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_BinaryValue)) __PYX_ERR(2, 89, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_BinaryValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_BinaryValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_BinaryValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_BinaryValue)) __PYX_ERR(2, 89, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_FieldValue = __Pyx_ImportType("thriftrw.wire.value", "FieldValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_FieldValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_FieldValue)) __PYX_ERR(2, 94, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_StructValue = __Pyx_ImportType("thriftrw.wire.value", "StructValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_StructValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_StructValue)) __PYX_ERR(2, 100, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_StructValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_StructValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_StructValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_StructValue)) __PYX_ERR(2, 100, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_MapItem = __Pyx_ImportType("thriftrw.wire.value", "MapItem", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_MapItem), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_MapItem)) __PYX_ERR(2, 105, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_MapValue = __Pyx_ImportType("thriftrw.wire.value", "MapValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_MapValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_MapValue)) __PYX_ERR(2, 110, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_MapValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_MapValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_MapValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_MapValue)) __PYX_ERR(2, 110, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_SetValue = __Pyx_ImportType("thriftrw.wire.value", "SetValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_SetValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_SetValue)) __PYX_ERR(2, 116, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_SetValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_SetValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_SetValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_SetValue)) __PYX_ERR(2, 116, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_5value_ListValue = __Pyx_ImportType("thriftrw.wire.value", "ListValue", sizeof(struct __pyx_obj_8thriftrw_4wire_5value_ListValue), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_5value_ListValue)) __PYX_ERR(2, 121, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_4wire_5value_ListValue = (struct __pyx_vtabstruct_8thriftrw_4wire_5value_ListValue*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_4wire_5value_ListValue->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_4wire_5value_ListValue)) __PYX_ERR(2, 121, __pyx_L1_error)
  __pyx_ptype_8thriftrw_4wire_7message_Message = __Pyx_ImportType("thriftrw.wire.message", "Message", sizeof(struct __pyx_obj_8thriftrw_4wire_7message_Message), 1); if (unlikely(!__pyx_ptype_8thriftrw_4wire_7message_Message)) __PYX_ERR(3, 26, __pyx_L1_error)
  __pyx_ptype_8thriftrw_7_buffer_ReadBuffer = __Pyx_ImportType("thriftrw._buffer", "ReadBuffer", sizeof(struct __pyx_obj_8thriftrw_7_buffer_ReadBuffer), 1); if (unlikely(!__pyx_ptype_8thriftrw_7_buffer_ReadBuffer)) __PYX_ERR(4, 24, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_7_buffer_ReadBuffer = (struct __pyx_vtabstruct_8thriftrw_7_buffer_ReadBuffer*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_7_buffer_ReadBuffer->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_7_buffer_ReadBuffer)) __PYX_ERR(4, 24, __pyx_L1_error)
  __pyx_ptype_8thriftrw_7_buffer_WriteBuffer = __Pyx_ImportType("thriftrw._buffer", "WriteBuffer", sizeof(struct __pyx_obj_8thriftrw_7_buffer_WriteBuffer), 1); if (unlikely(!__pyx_ptype_8thriftrw_7_buffer_WriteBuffer)) __PYX_ERR(4, 39, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_7_buffer_WriteBuffer = (struct __pyx_vtabstruct_8thriftrw_7_buffer_WriteBuffer*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_7_buffer_WriteBuffer->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_7_buffer_WriteBuffer)) __PYX_ERR(4, 39, __pyx_L1_error)
  __pyx_ptype_8thriftrw_8protocol_4core_MessageHeader = __Pyx_ImportType("thriftrw.protocol.core", "MessageHeader", sizeof(struct __pyx_obj_8thriftrw_8protocol_4core_MessageHeader), 1); if (unlikely(!__pyx_ptype_8thriftrw_8protocol_4core_MessageHeader)) __PYX_ERR(5, 57, __pyx_L1_error)
  __pyx_ptype_8thriftrw_8protocol_4core__ValueWriter = __Pyx_ImportType("thriftrw.protocol.core", "_ValueWriter", sizeof(struct __pyx_obj_8thriftrw_8protocol_4core__ValueWriter), 1); if (unlikely(!__pyx_ptype_8thriftrw_8protocol_4core__ValueWriter)) __PYX_ERR(5, 160, __pyx_L1_error)
  __pyx_vtabptr_8thriftrw_8protocol_4core__ValueWriter = (struct __pyx_vtabstruct_8thriftrw_8protocol_4core__ValueWriter*)__Pyx_GetVtable(__pyx_ptype_8thriftrw_8protocol_4core__ValueWriter->tp_dict); if (unlikely(!__pyx_vtabptr_8thriftrw_8protocol_4core__ValueWriter)) __PYX_ERR(5, 160, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initbinary(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initbinary(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_binary(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_binary(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec_binary(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_binary(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("binary", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_thriftrw__protocol__binary) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "thriftrw.protocol.binary")) {
      if (unlikely(PyDict_SetItemString(modules, "thriftrw.protocol.binary", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "thriftrw/protocol/binary.pyx":63
 *     MessageHeader,
 * )
 * from ..errors import EndOfInputError             # <<<<<<<<<<<<<<
 * from ..errors import ThriftProtocolError
 * from ._endian cimport (
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_EndOfInputError);
  __Pyx_GIVEREF(__pyx_n_s_EndOfInputError);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_EndOfInputError);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_errors, __pyx_t_1, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_EndOfInputError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EndOfInputError, __pyx_t_1) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "thriftrw/protocol/binary.pyx":64
 * )
 * from ..errors import EndOfInputError
 * from ..errors import ThriftProtocolError             # <<<<<<<<<<<<<<
 * from ._endian cimport (
 *     htobe16,
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ThriftProtocolError);
  __Pyx_GIVEREF(__pyx_n_s_ThriftProtocolError);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ThriftProtocolError);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_errors, __pyx_t_2, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ThriftProtocolError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ThriftProtocolError, __pyx_t_2) < 0) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "thriftrw/protocol/binary.pyx":75
 * 
 * 
 * cdef STRUCT_END_HEADER = FieldHeader(-1, -1)             # <<<<<<<<<<<<<<
 * 
 * cdef int8_t STRUCT_END = 0
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_int_neg_1) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_id, __pyx_int_neg_1) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_8thriftrw_8protocol_6binary_STRUCT_END_HEADER);
  __Pyx_DECREF_SET(__pyx_v_8thriftrw_8protocol_6binary_STRUCT_END_HEADER, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "thriftrw/protocol/binary.pyx":77
 * cdef STRUCT_END_HEADER = FieldHeader(-1, -1)
 * 
 * cdef int8_t STRUCT_END = 0             # <<<<<<<<<<<<<<
 * 
 * cdef int32_t VERSION = 1
 */
  __pyx_v_8thriftrw_8protocol_6binary_STRUCT_END = 0;

  /* "thriftrw/protocol/binary.pyx":79
 * cdef int8_t STRUCT_END = 0
 * 
 * cdef int32_t VERSION = 1             # <<<<<<<<<<<<<<
 * 
 * # Under strict mode, the version number is in most significant 16-bits of the
 */
  __pyx_v_8thriftrw_8protocol_6binary_VERSION = 1;

  /* "thriftrw/protocol/binary.pyx":85
 * # with the integer to get just the version number (it still needs to be
 * # shifted 16-bits to the right).
 * cdef int32_t VERSION_MASK = 0x7fff0000             # <<<<<<<<<<<<<<
 * 
 * # The least significant 8-bits of the 32-bit integer contain the type.
 */
  __pyx_v_8thriftrw_8protocol_6binary_VERSION_MASK = 0x7fff0000;

  /* "thriftrw/protocol/binary.pyx":88
 * 
 * # The least significant 8-bits of the 32-bit integer contain the type.
 * cdef int32_t TYPE_MASK = 0x000000ff             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_8thriftrw_8protocol_6binary_TYPE_MASK = 0x000000ff;

  /* "thriftrw/protocol/binary.pyx":380
 * 
 * 
 * __all__ = ['BinaryProtocol']             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_BinaryProtocol);
  __Pyx_GIVEREF(__pyx_n_u_BinaryProtocol);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_BinaryProtocol);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_1) < 0) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_BinaryProtocol(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_8thriftrw_8protocol_6binary_1__pyx_unpickle_BinaryProtocol, NULL, __pyx_n_s_thriftrw_protocol_binary); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_BinaryProtocol, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "thriftrw/protocol/binary.pyx":1
 * # Copyright (c) 2016 Uber Technologies, Inc.             # <<<<<<<<<<<<<<
 * #
 * # Permission is hereby granted, free of charge, to any person obtaining a copy
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py")
 * cdef object __Pyx_CFunc_ListValue______OldBinaryProtocolReader___to_py(ListValue (*f)(_OldBinaryProtocolReader) ):             # <<<<<<<<<<<<<<
 *     def wrap(_OldBinaryProtocolReader self):
 *         """wrap(self: '_OldBinaryProtocolReader') -> 'ListValue'"""
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init thriftrw.protocol.binary", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init thriftrw.protocol.binary");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    if (likely(result)) {
        Py_INCREF(result);
    } else if (unlikely(PyErr_Occurred())) {
        result = NULL;
    } else {
#else
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyCFunctionFastCall */
    #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
    #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
    #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* ExtTypeTest */
    static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
    static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
    static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* Import */
    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
    static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
    static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetItemInt */
    static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely((0 <= wrapped_i) & (wrapped_i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely((0 <= wrapped_i) & (wrapped_i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* DictGetItem */
    #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SaveResetException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if PY_VERSION_HEX >= 0x030700A3
    *type = tstate->exc_state.exc_type;
    *value = tstate->exc_state.exc_value;
    *tb = tstate->exc_state.exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = type;
    tstate->exc_state.exc_value = value;
    tstate->exc_state.exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = local_type;
    tstate->exc_state.exc_value = local_value;
    tstate->exc_state.exc_traceback = local_tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* FetchCommonType */
      static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
      #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CallNextTpDealloc */
          static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
          static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
          static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* GetVTable */
          static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* PyObject_GenericGetAttrNoDict */
          #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
          #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
          static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
          static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* CLineInTraceback */
          #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
          static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
          #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
          #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int8_t(int8_t value) {
    const int8_t neg_one = (int8_t) -1, const_zero = (int8_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int8_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int16_t(int16_t value) {
    const int16_t neg_one = (int16_t) -1, const_zero = (int16_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int16_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int16_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int16_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
          static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE int8_t __Pyx_PyInt_As_int8_t(PyObject *x) {
    const int8_t neg_one = (int8_t) -1, const_zero = (int8_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 2 * PyLong_SHIFT) {
                            return (int8_t) (((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 3 * PyLong_SHIFT) {
                            return (int8_t) (((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) >= 4 * PyLong_SHIFT) {
                            return (int8_t) (((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int8_t) ((((((int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int8_t) ((((((((int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int8_t) (((int8_t)-1)*(((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int8_t) ((((((((((int8_t)digits[3]) << PyLong_SHIFT) | (int8_t)digits[2]) << PyLong_SHIFT) | (int8_t)digits[1]) << PyLong_SHIFT) | (int8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int8_t) -1;
        }
    } else {
        int8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int8_t) -1;
        val = __Pyx_PyInt_As_int8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int8_t");
    return (int8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int8_t");
    return (int8_t) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *x) {
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 2 * PyLong_SHIFT) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 3 * PyLong_SHIFT) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 4 * PyLong_SHIFT) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int32_t) -1;
        }
    } else {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyInt_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE int16_t __Pyx_PyInt_As_int16_t(PyObject *x) {
    const int16_t neg_one = (int16_t) -1, const_zero = (int16_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int16_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int16_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int16_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 2 * PyLong_SHIFT) {
                            return (int16_t) (((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 3 * PyLong_SHIFT) {
                            return (int16_t) (((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) >= 4 * PyLong_SHIFT) {
                            return (int16_t) (((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int16_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int16_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int16_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int16_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int16_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int16_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int16_t) ((((((int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int16_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int16_t) ((((((((int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int16_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int16_t) (((int16_t)-1)*(((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int16_t) ((((((((((int16_t)digits[3]) << PyLong_SHIFT) | (int16_t)digits[2]) << PyLong_SHIFT) | (int16_t)digits[1]) << PyLong_SHIFT) | (int16_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int16_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int16_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int16_t) -1;
        }
    } else {
        int16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int16_t) -1;
        val = __Pyx_PyInt_As_int16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int16_t");
    return (int16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int16_t");
    return (int16_t) -1;
}

/* FastTypeChecks */
          #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
          static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* ModuleImport */
          #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
          #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* InitStrings */
          static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
